from datetime import datetime
from unittest.mock import Mock, call, patch

import pytest

from oumi.core.configs import JobConfig, JobResources, StorageMount
from oumi.core.launcher import JobState, JobStatus
from oumi.launcher.clients.polaris_client import PolarisClient
from oumi.launcher.clusters.polaris_cluster import PolarisCluster


#
# Fixtures
#
@pytest.fixture
def mock_polaris_client():
    yield Mock(spec=PolarisClient)


@pytest.fixture
def mock_datetime():
    with patch("oumi.launcher.clusters.polaris_cluster.datetime") as mock_dt:
        mock_dt.now.return_value = datetime(2024, 10, 9, 13, 4, 24, 513094)
        yield mock_dt


def _get_default_job(cloud: str) -> JobConfig:
    resources = JobResources(
        cloud=cloud,
        region="us-central1",
        zone=None,
        accelerators="A100-80GB",
        cpus="4",
        memory="64",
        instance_type=None,
        use_spot=True,
        disk_size=512,
        disk_tier="low",
    )
    return JobConfig(
        name="myjob",
        user="user",
        working_dir="./",
        num_nodes=2,
        resources=resources,
        envs={"var1": "val1"},
        file_mounts={
            "~/home/remote/path.bar": "~/local/path.bar",
            "~/home/remote/path2.txt": "~/local/path2.txt",
        },
        storage_mounts={
            "~/home/remote/path/gcs/": StorageMount(
                source="gs://mybucket/", store="gcs"
            )
        },
        setup=(
            "#PBS -o some/log \n#PBE -l wow\n#PBS -e run/log\n"
            "pip install -r requirements.txt"
        ),
        run="./hello_world.sh",
    )


#
# Tests
#
def test_polaris_cluster_name(mock_datetime, mock_polaris_client):
    cluster = PolarisCluster("demand.einstein", mock_polaris_client)
    assert cluster.name() == "demand.einstein"

    cluster = PolarisCluster("debug.einstein", mock_polaris_client)
    assert cluster.name() == "debug.einstein"

    cluster = PolarisCluster("debug-scaling.einstein", mock_polaris_client)
    assert cluster.name() == "debug-scaling.einstein"

    cluster = PolarisCluster("preemptable.einstein", mock_polaris_client)
    assert cluster.name() == "preemptable.einstein"

    cluster = PolarisCluster("prod.einstein", mock_polaris_client)
    assert cluster.name() == "prod.einstein"


def test_polaris_cluster_invalid_name(mock_datetime, mock_polaris_client):
    with pytest.raises(ValueError):
        PolarisCluster("einstein", mock_polaris_client)


def test_polaris_cluster_invalid_queue(mock_datetime, mock_polaris_client):
    with pytest.raises(ValueError):
        PolarisCluster("albert.einstein", mock_polaris_client)


def test_polaris_cluster_get_job_valid_id(mock_datetime, mock_polaris_client):
    cluster = PolarisCluster("debug.name", mock_polaris_client)
    mock_polaris_client.list_jobs.return_value = [
        JobStatus(
            id="myjob",
            name="some name",
            status="running",
            metadata="",
            cluster="mycluster",
            done=False,
            state=JobState.PENDING,
        ),
        JobStatus(
            id="job2",
            name="some",
            status="running",
            metadata="",
            cluster="mycluster",
            done=False,
            state=JobState.PENDING,
        ),
        JobStatus(
            id="final job",
            name="name3",
            status="running",
            metadata="",
            cluster="mycluster",
            done=False,
            state=JobState.PENDING,
        ),
    ]
    job = cluster.get_job("myjob")
    mock_polaris_client.list_jobs.assert_called_once_with(
        PolarisClient.SupportedQueues.DEBUG
    )
    assert job is not None
    assert job.id == "myjob"
    assert job.cluster == "debug.name"


def test_polaris_cluster_get_job_invalid_id_empty(mock_datetime, mock_polaris_client):
    cluster = PolarisCluster("debug.name", mock_polaris_client)
    mock_polaris_client.list_jobs.return_value = []
    job = cluster.get_job("myjob")
    mock_polaris_client.list_jobs.assert_called_once_with(
        PolarisClient.SupportedQueues.DEBUG
    )
    assert job is None


def test_polaris_cluster_get_job_invalid_id_nonempty(
    mock_datetime, mock_polaris_client
):
    cluster = PolarisCluster("debug.name", mock_polaris_client)
    mock_polaris_client.list_jobs.return_value = [
        JobStatus(
            id="myjob",
            name="some name",
            status="running",
            metadata="",
            cluster="mycluster",
            done=False,
            state=JobState.PENDING,
        ),
        JobStatus(
            id="job2",
            name="some",
            status="running",
            metadata="",
            cluster="mycluster",
            done=False,
            state=JobState.PENDING,
        ),
        JobStatus(
            id="final job",
            name="name3",
            status="running",
            metadata="",
            cluster="mycluster",
            done=False,
            state=JobState.PENDING,
        ),
    ]
    job = cluster.get_job("wrong job")
    mock_polaris_client.list_jobs.assert_called_once_with(
        PolarisClient.SupportedQueues.DEBUG
    )
    assert job is None


def test_polaris_cluster_get_jobs_nonempty(mock_datetime, mock_polaris_client):
    cluster = PolarisCluster("debug.name", mock_polaris_client)
    mock_polaris_client.list_jobs.return_value = [
        JobStatus(
            id="myjob",
            name="some name",
            status="running",
            metadata="",
            cluster="mycluster",
            done=False,
            state=JobState.PENDING,
        ),
        JobStatus(
            id="job2",
            name="some",
            status="running",
            metadata="",
            cluster="mycluster",
            done=False,
            state=JobState.PENDING,
        ),
        JobStatus(
            id="final job",
            name="name3",
            status="running",
            metadata="",
            cluster="mycluster",
            done=False,
            state=JobState.PENDING,
        ),
    ]
    jobs = cluster.get_jobs()
    mock_polaris_client.list_jobs.assert_called_once_with(
        PolarisClient.SupportedQueues.DEBUG
    )
    expected_jobs = [
        JobStatus(
            id="myjob",
            name="some name",
            status="running",
            metadata="",
            cluster="debug.name",
            done=False,
            state=JobState.PENDING,
        ),
        JobStatus(
            id="job2",
            name="some",
            status="running",
            metadata="",
            cluster="debug.name",
            done=False,
            state=JobState.PENDING,
        ),
        JobStatus(
            id="final job",
            name="name3",
            status="running",
            metadata="",
            cluster="debug.name",
            done=False,
            state=JobState.PENDING,
        ),
    ]
    assert jobs == expected_jobs


def test_polaris_cluster_get_jobs_empty(mock_datetime, mock_polaris_client):
    cluster = PolarisCluster("debug.name", mock_polaris_client)
    mock_polaris_client.list_jobs.return_value = []
    jobs = cluster.get_jobs()
    mock_polaris_client.list_jobs.assert_called_once_with(
        PolarisClient.SupportedQueues.DEBUG
    )
    expected_jobs = []
    assert jobs == expected_jobs


def test_polaris_cluster_cancel_job(mock_datetime, mock_polaris_client):
    cluster = PolarisCluster("prod.name", mock_polaris_client)
    mock_polaris_client.list_jobs.return_value = [
        JobStatus(
            id="myjob",
            name="some name",
            status="running",
            metadata="",
            cluster="debug.name",
            done=False,
            state=JobState.PENDING,
        ),
        JobStatus(
            id="job2",
            name="some",
            status="running",
            metadata="",
            cluster="debug.name",
            done=False,
            state=JobState.PENDING,
        ),
        JobStatus(
            id="final job",
            name="name3",
            status="running",
            metadata="",
            cluster="debug.name",
            done=False,
            state=JobState.PENDING,
        ),
    ]
    job_status = cluster.cancel_job("job2")
    expected_status = JobStatus(
        id="job2",
        name="some",
        status="running",
        metadata="",
        cluster="prod.name",
        done=False,
        state=JobState.PENDING,
    )
    mock_polaris_client.cancel.assert_called_once_with(
        "job2",
        PolarisClient.SupportedQueues.PROD,
    )
    assert job_status == expected_status


def test_polaris_cluster_cancel_job_fails(mock_datetime, mock_polaris_client):
    cluster = PolarisCluster("prod.name", mock_polaris_client)
    mock_polaris_client.list_jobs.return_value = [
        JobStatus(
            id="job2",
            name="some",
            status="running",
            metadata="",
            cluster="debug.name",
            done=False,
            state=JobState.PENDING,
        ),
    ]
    with pytest.raises(RuntimeError):
        _ = cluster.cancel_job("myjobid")


def test_polaris_cluster_run_job(mock_datetime, mock_polaris_client):
    cluster = PolarisCluster("debug.name", mock_polaris_client)
    mock_successful_cmd = Mock()
    mock_successful_cmd.exit_code = 0
    mock_polaris_client.run_commands.return_value = mock_successful_cmd
    mock_polaris_client.submit_job.return_value = "1234"
    mock_polaris_client.list_jobs.return_value = [
        JobStatus(
            id="1234",
            name="some name",
            status="queued",
            metadata="",
            cluster="mycluster",
            done=False,
            state=JobState.PENDING,
        )
    ]
    expected_status = JobStatus(
        id="1234",
        name="some name",
        status="queued",
        metadata="",
        cluster="debug.name",
        done=False,
        state=JobState.PENDING,
    )
    job_status = cluster.run_job(_get_default_job("polaris"))
    mock_polaris_client.put_recursive.assert_has_calls(
        [
            call(
                "./",
                "/home/user/oumi_launcher/20241009_130424513094",
            ),
            call(
                "~/local/path.bar",
                "~/home/remote/path.bar",
            ),
            call(
                "~/local/path2.txt",
                "~/home/remote/path2.txt",
            ),
        ],
    )
    mock_polaris_client.run_commands.assert_has_calls(
        [
            call(
                [
                    "cd /home/user/oumi_launcher/20241009_130424513094",
                    "module use /soft/modulefiles",
                    "module load conda",
                    "if [ ! -d /home/$USER/miniconda3/envs/oumi ]; then",
                    'echo "Creating Oumi Conda environment... '
                    '---------------------------"',
                    "conda create -y python=3.11 --prefix "
                    "/home/$USER/miniconda3/envs/oumi",
                    "fi",
                    'echo "Installing packages... '
                    '---------------------------------------"',
                    "conda activate /home/$USER/miniconda3/envs/oumi",
                    "if ! command -v uv >/dev/null 2>&1; then",
                    "pip install -U uv",
                    "fi",
                    "pip install -e '.[gpu]'",
                ]
            ),
            call(
                ["chmod +x /home/user/oumi_launcher/20241009_130424513094/oumi_job.sh"]
            ),
            call(
                [
                    "mkdir -p some/log",
                    "mkdir -p run/log",
                ]
            ),
        ]
    )
    job_script = (
        "#!/bin/bash\n#PBS -o some/log \n#PBE -l wow\n#PBS -e run/log\n\n"
        "export var1=val1\n\n"
        "pip install -r requirements.txt\n./hello_world.sh\n"
    )
    mock_polaris_client.put.assert_called_once_with(
        job_script, "/home/user/oumi_launcher/20241009_130424513094/oumi_job.sh"
    )
    mock_polaris_client.submit_job.assert_called_once_with(
        "/home/user/oumi_launcher/20241009_130424513094/oumi_job.sh",
        "/home/user/oumi_launcher/20241009_130424513094",
        2,
        PolarisClient.SupportedQueues.DEBUG,
        "myjob",
    )
    mock_polaris_client.list_jobs.assert_called_once_with(
        PolarisClient.SupportedQueues.DEBUG
    )
    assert job_status == expected_status


def test_polaris_cluster_run_job_no_working_dir(mock_datetime, mock_polaris_client):
    cluster = PolarisCluster("debug.name", mock_polaris_client)
    mock_successful_cmd = Mock()
    mock_successful_cmd.exit_code = 0
    mock_polaris_client.run_commands.return_value = mock_successful_cmd
    mock_polaris_client.submit_job.return_value = "1234"
    mock_polaris_client.list_jobs.return_value = [
        JobStatus(
            id="1234",
            name="some name",
            status="queued",
            metadata="",
            cluster="mycluster",
            done=False,
            state=JobState.PENDING,
        )
    ]
    expected_status = JobStatus(
        id="1234",
        name="some name",
        status="queued",
        metadata="",
        cluster="debug.name",
        done=False,
        state=JobState.PENDING,
    )
    job_config = _get_default_job("polaris")
    job_config.working_dir = None
    job_config.file_mounts = {}
    job_status = cluster.run_job(job_config)
    mock_polaris_client.put_recursive.assert_not_called()
    mock_polaris_client.run_commands.assert_has_calls(
        [
            call(["mkdir -p /home/user/oumi_launcher/20241009_130424513094"]),
            call(
                [
                    "cd /home/user/oumi_launcher/20241009_130424513094",
                    "module use /soft/modulefiles",
                    "module load conda",
                    "if [ ! -d /home/$USER/miniconda3/envs/oumi ]; then",
                    'echo "Creating Oumi Conda environment... '
                    '---------------------------"',
                    "conda create -y python=3.11 --prefix "
                    "/home/$USER/miniconda3/envs/oumi",
                    "fi",
                    'echo "Installing packages... '
                    '---------------------------------------"',
                    "conda activate /home/$USER/miniconda3/envs/oumi",
                    "if ! command -v uv >/dev/null 2>&1; then",
                    "pip install -U uv",
                    "fi",
                    "pip install -e '.[gpu]'",
                ],
            ),
            call(
                ["chmod +x /home/user/oumi_launcher/20241009_130424513094/oumi_job.sh"]
            ),
            call(
                [
                    "mkdir -p some/log",
                    "mkdir -p run/log",
                ]
            ),
        ]
    )
    job_script = (
        "#!/bin/bash\n#PBS -o some/log \n#PBE -l wow\n#PBS -e run/log\n\n"
        "export var1=val1\n\n"
        "pip install -r requirements.txt\n./hello_world.sh\n"
    )
    mock_polaris_client.put.assert_called_once_with(
        job_script, "/home/user/oumi_launcher/20241009_130424513094/oumi_job.sh"
    )
    mock_polaris_client.submit_job.assert_called_once_with(
        "/home/user/oumi_launcher/20241009_130424513094/oumi_job.sh",
        "/home/user/oumi_launcher/20241009_130424513094",
        2,
        PolarisClient.SupportedQueues.DEBUG,
        "myjob",
    )
    mock_polaris_client.list_jobs.assert_called_once_with(
        PolarisClient.SupportedQueues.DEBUG
    )
    assert job_status == expected_status


def test_polaris_cluster_run_job_with_conda_setup(mock_datetime, mock_polaris_client):
    mock_successful_cmd = Mock()
    mock_successful_cmd.exit_code = 0
    mock_failed_cmd = Mock()
    mock_failed_cmd.exit_code = 1
    mock_polaris_client.run_commands.side_effect = [
        mock_failed_cmd,
        mock_successful_cmd,
        mock_successful_cmd,
        mock_successful_cmd,
    ]
    cluster = PolarisCluster("debug.name", mock_polaris_client)
    mock_polaris_client.submit_job.return_value = "1234"
    mock_polaris_client.list_jobs.return_value = [
        JobStatus(
            id="1234",
            name="some name",
            status="queued",
            metadata="",
            cluster="mycluster",
            done=False,
            state=JobState.PENDING,
        )
    ]
    expected_status = JobStatus(
        id="1234",
        name="some name",
        status="queued",
        metadata="",
        cluster="debug.name",
        done=False,
        state=JobState.PENDING,
    )
    job_status = cluster.run_job(_get_default_job("polaris"))
    mock_polaris_client.put_recursive.assert_has_calls(
        [
            call(
                "./",
                "/home/user/oumi_launcher/20241009_130424513094",
            ),
            call(
                "~/local/path.bar",
                "~/home/remote/path.bar",
            ),
            call(
                "~/local/path2.txt",
                "~/home/remote/path2.txt",
            ),
        ],
    )
    mock_polaris_client.run_commands.assert_has_calls(
        [
            call(
                [
                    "cd /home/user/oumi_launcher/20241009_130424513094",
                    "module use /soft/modulefiles",
                    "module load conda",
                    "if [ ! -d /home/$USER/miniconda3/envs/oumi ]; then",
                    'echo "Creating Oumi Conda environment... '
                    '---------------------------"',
                    "conda create -y python=3.11 --prefix "
                    "/home/$USER/miniconda3/envs/oumi",
                    "fi",
                    'echo "Installing packages... '
                    '---------------------------------------"',
                    "conda activate /home/$USER/miniconda3/envs/oumi",
                    "if ! command -v uv >/dev/null 2>&1; then",
                    "pip install -U uv",
                    "fi",
                    "pip install -e '.[gpu]'",
                ]
            ),
            call(
                ["chmod +x /home/user/oumi_launcher/20241009_130424513094/oumi_job.sh"]
            ),
            call(
                [
                    "mkdir -p some/log",
                    "mkdir -p run/log",
                ]
            ),
        ]
    )
    job_script = (
        "#!/bin/bash\n#PBS -o some/log \n#PBE -l wow\n#PBS -e run/log\n\n"
        "export var1=val1\n\n"
        "pip install -r requirements.txt\n./hello_world.sh\n"
    )
    mock_polaris_client.put.assert_called_once_with(
        job_script, "/home/user/oumi_launcher/20241009_130424513094/oumi_job.sh"
    )
    mock_polaris_client.submit_job.assert_called_once_with(
        "/home/user/oumi_launcher/20241009_130424513094/oumi_job.sh",
        "/home/user/oumi_launcher/20241009_130424513094",
        2,
        PolarisClient.SupportedQueues.DEBUG,
        "myjob",
    )
    mock_polaris_client.list_jobs.assert_called_once_with(
        PolarisClient.SupportedQueues.DEBUG
    )
    assert job_status == expected_status


def test_polaris_cluster_run_job_no_name(mock_datetime, mock_polaris_client):
    mock_successful_cmd = Mock()
    mock_successful_cmd.exit_code = 0
    mock_polaris_client.run_commands.return_value = mock_successful_cmd
    cluster = PolarisCluster("debug.name", mock_polaris_client)
    mock_polaris_client.submit_job.return_value = "1234"
    mock_polaris_client.list_jobs.return_value = [
        JobStatus(
            id="1234",
            name="some name",
            status="queued",
            metadata="",
            cluster="mycluster",
            done=False,
            state=JobState.PENDING,
        )
    ]
    expected_status = JobStatus(
        id="1234",
        name="some name",
        status="queued",
        metadata="",
        cluster="debug.name",
        done=False,
        state=JobState.PENDING,
    )
    job = _get_default_job("polaris")
    job.name = None
    with patch("oumi.launcher.clusters.polaris_cluster.uuid") as mock_uuid:
        mock_hex = Mock()
        mock_hex.hex = "1-2-3"
        mock_uuid.uuid1.return_value = mock_hex
        job_status = cluster.run_job(job)
    mock_polaris_client.put_recursive.assert_has_calls(
        [
            call(
                "./",
                "/home/user/oumi_launcher/20241009_130424513094",
            ),
            call(
                "~/local/path.bar",
                "~/home/remote/path.bar",
            ),
            call(
                "~/local/path2.txt",
                "~/home/remote/path2.txt",
            ),
        ],
    )
    mock_polaris_client.run_commands.assert_has_calls(
        [
            call(
                [
                    "cd /home/user/oumi_launcher/20241009_130424513094",
                    "module use /soft/modulefiles",
                    "module load conda",
                    "if [ ! -d /home/$USER/miniconda3/envs/oumi ]; then",
                    'echo "Creating Oumi Conda environment... '
                    '---------------------------"',
                    "conda create -y python=3.11 --prefix "
                    "/home/$USER/miniconda3/envs/oumi",
                    "fi",
                    'echo "Installing packages... '
                    '---------------------------------------"',
                    "conda activate /home/$USER/miniconda3/envs/oumi",
                    "if ! command -v uv >/dev/null 2>&1; then",
                    "pip install -U uv",
                    "fi",
                    "pip install -e '.[gpu]'",
                ]
            ),
            call(
                ["chmod +x /home/user/oumi_launcher/20241009_130424513094/oumi_job.sh"]
            ),
            call(
                [
                    "mkdir -p some/log",
                    "mkdir -p run/log",
                ]
            ),
        ]
    )
    job_script = (
        "#!/bin/bash\n#PBS -o some/log \n#PBE -l wow\n#PBS -e run/log\n\n"
        "export var1=val1\n\n"
        "pip install -r requirements.txt\n./hello_world.sh\n"
    )
    mock_polaris_client.put.assert_called_once_with(
        job_script, "/home/user/oumi_launcher/20241009_130424513094/oumi_job.sh"
    )
    mock_polaris_client.submit_job.assert_called_once_with(
        "/home/user/oumi_launcher/20241009_130424513094/oumi_job.sh",
        "/home/user/oumi_launcher/20241009_130424513094",
        2,
        PolarisClient.SupportedQueues.DEBUG,
        "1-2-3",
    )
    mock_polaris_client.list_jobs.assert_called_once_with(
        PolarisClient.SupportedQueues.DEBUG
    )
    assert job_status == expected_status


def test_polaris_cluster_run_job_no_mounts(mock_datetime, mock_polaris_client):
    mock_successful_cmd = Mock()
    mock_successful_cmd.exit_code = 0
    mock_polaris_client.run_commands.return_value = mock_successful_cmd
    cluster = PolarisCluster("debug.name", mock_polaris_client)
    mock_polaris_client.submit_job.return_value = "1234"
    mock_polaris_client.list_jobs.return_value = [
        JobStatus(
            id="1234",
            name="some name",
            status="queued",
            metadata="",
            cluster="mycluster",
            done=False,
            state=JobState.PENDING,
        )
    ]
    expected_status = JobStatus(
        id="1234",
        name="some name",
        status="queued",
        metadata="",
        cluster="debug.name",
        done=False,
        state=JobState.PENDING,
    )
    job = _get_default_job("polaris")
    job.file_mounts = {}
    job_status = cluster.run_job(job)
    mock_polaris_client.put_recursive.assert_has_calls(
        [
            call(
                "./",
                "/home/user/oumi_launcher/20241009_130424513094",
            ),
        ],
    )
    mock_polaris_client.run_commands.assert_has_calls(
        [
            call(
                [
                    "cd /home/user/oumi_launcher/20241009_130424513094",
                    "module use /soft/modulefiles",
                    "module load conda",
                    "if [ ! -d /home/$USER/miniconda3/envs/oumi ]; then",
                    'echo "Creating Oumi Conda environment... '
                    '---------------------------"',
                    "conda create -y python=3.11 --prefix "
                    "/home/$USER/miniconda3/envs/oumi",
                    "fi",
                    'echo "Installing packages... '
                    '---------------------------------------"',
                    "conda activate /home/$USER/miniconda3/envs/oumi",
                    "if ! command -v uv >/dev/null 2>&1; then",
                    "pip install -U uv",
                    "fi",
                    "pip install -e '.[gpu]'",
                ]
            ),
            call(
                ["chmod +x /home/user/oumi_launcher/20241009_130424513094/oumi_job.sh"]
            ),
            call(
                [
                    "mkdir -p some/log",
                    "mkdir -p run/log",
                ]
            ),
        ]
    )
    job_script = (
        "#!/bin/bash\n#PBS -o some/log \n#PBE -l wow\n#PBS -e run/log\n\n"
        "export var1=val1\n\n"
        "pip install -r requirements.txt\n./hello_world.sh\n"
    )
    mock_polaris_client.put.assert_called_once_with(
        job_script, "/home/user/oumi_launcher/20241009_130424513094/oumi_job.sh"
    )
    mock_polaris_client.submit_job.assert_called_once_with(
        "/home/user/oumi_launcher/20241009_130424513094/oumi_job.sh",
        "/home/user/oumi_launcher/20241009_130424513094",
        2,
        PolarisClient.SupportedQueues.DEBUG,
        "myjob",
    )
    mock_polaris_client.list_jobs.assert_called_once_with(
        PolarisClient.SupportedQueues.DEBUG
    )
    assert job_status == expected_status


def test_polaris_cluster_run_job_no_pbs(mock_datetime, mock_polaris_client):
    mock_successful_cmd = Mock()
    mock_successful_cmd.exit_code = 0
    mock_polaris_client.run_commands.return_value = mock_successful_cmd
    cluster = PolarisCluster("debug.name", mock_polaris_client)
    mock_polaris_client.submit_job.return_value = "1234"
    mock_polaris_client.list_jobs.return_value = [
        JobStatus(
            id="1234",
            name="some name",
            status="queued",
            metadata="",
            cluster="mycluster",
            done=False,
            state=JobState.PENDING,
        )
    ]
    expected_status = JobStatus(
        id="1234",
        name="some name",
        status="queued",
        metadata="",
        cluster="debug.name",
        done=False,
        state=JobState.PENDING,
    )
    job = _get_default_job("polaris")
    job.file_mounts = {}
    job.setup = "small setup"
    job.run = "./hello_world.sh"
    job_status = cluster.run_job(job)
    mock_polaris_client.put_recursive.assert_has_calls(
        [
            call(
                "./",
                "/home/user/oumi_launcher/20241009_130424513094",
            ),
        ],
    )
    mock_polaris_client.run_commands.assert_has_calls(
        [
            call(
                [
                    "cd /home/user/oumi_launcher/20241009_130424513094",
                    "module use /soft/modulefiles",
                    "module load conda",
                    "if [ ! -d /home/$USER/miniconda3/envs/oumi ]; then",
                    'echo "Creating Oumi Conda environment... '
                    '---------------------------"',
                    "conda create -y python=3.11 --prefix "
                    "/home/$USER/miniconda3/envs/oumi",
                    "fi",
                    'echo "Installing packages... '
                    '---------------------------------------"',
                    "conda activate /home/$USER/miniconda3/envs/oumi",
                    "if ! command -v uv >/dev/null 2>&1; then",
                    "pip install -U uv",
                    "fi",
                    "pip install -e '.[gpu]'",
                ]
            ),
            call(
                ["chmod +x /home/user/oumi_launcher/20241009_130424513094/oumi_job.sh"]
            ),
        ]
    )
    job_script = "#!/bin/bash\n\nexport var1=val1\n\nsmall setup\n./hello_world.sh\n"
    mock_polaris_client.put.assert_called_once_with(
        job_script, "/home/user/oumi_launcher/20241009_130424513094/oumi_job.sh"
    )
    mock_polaris_client.submit_job.assert_called_once_with(
        "/home/user/oumi_launcher/20241009_130424513094/oumi_job.sh",
        "/home/user/oumi_launcher/20241009_130424513094",
        2,
        PolarisClient.SupportedQueues.DEBUG,
        "myjob",
    )
    mock_polaris_client.list_jobs.assert_called_once_with(
        PolarisClient.SupportedQueues.DEBUG
    )
    assert job_status == expected_status


def test_polaris_cluster_run_job_no_setup(mock_datetime, mock_polaris_client):
    mock_successful_cmd = Mock()
    mock_successful_cmd.exit_code = 0
    mock_polaris_client.run_commands.return_value = mock_successful_cmd
    cluster = PolarisCluster("debug.name", mock_polaris_client)
    mock_polaris_client.submit_job.return_value = "1234"
    mock_polaris_client.list_jobs.return_value = [
        JobStatus(
            id="1234",
            name="some name",
            status="queued",
            metadata="",
            cluster="mycluster",
            done=False,
            state=JobState.PENDING,
        )
    ]
    expected_status = JobStatus(
        id="1234",
        name="some name",
        status="queued",
        metadata="",
        cluster="debug.name",
        done=False,
        state=JobState.PENDING,
    )
    job = _get_default_job("polaris")
    job.file_mounts = {}
    job.setup = None
    job.run = "./hello_world.sh"
    job_status = cluster.run_job(job)
    mock_polaris_client.put_recursive.assert_has_calls(
        [
            call(
                "./",
                "/home/user/oumi_launcher/20241009_130424513094",
            ),
        ],
    )
    mock_polaris_client.run_commands.assert_has_calls(
        [
            call(
                [
                    "cd /home/user/oumi_launcher/20241009_130424513094",
                    "module use /soft/modulefiles",
                    "module load conda",
                    "if [ ! -d /home/$USER/miniconda3/envs/oumi ]; then",
                    'echo "Creating Oumi Conda environment... '
                    '---------------------------"',
                    "conda create -y python=3.11 --prefix "
                    "/home/$USER/miniconda3/envs/oumi",
                    "fi",
                    'echo "Installing packages... '
                    '---------------------------------------"',
                    "conda activate /home/$USER/miniconda3/envs/oumi",
                    "if ! command -v uv >/dev/null 2>&1; then",
                    "pip install -U uv",
                    "fi",
                    "pip install -e '.[gpu]'",
                ]
            ),
            call(
                ["chmod +x /home/user/oumi_launcher/20241009_130424513094/oumi_job.sh"]
            ),
        ]
    )
    job_script = "#!/bin/bash\n\nexport var1=val1\n\n./hello_world.sh\n"
    mock_polaris_client.put.assert_called_once_with(
        job_script, "/home/user/oumi_launcher/20241009_130424513094/oumi_job.sh"
    )
    mock_polaris_client.submit_job.assert_called_once_with(
        "/home/user/oumi_launcher/20241009_130424513094/oumi_job.sh",
        "/home/user/oumi_launcher/20241009_130424513094",
        2,
        PolarisClient.SupportedQueues.DEBUG,
        "myjob",
    )
    mock_polaris_client.list_jobs.assert_called_once_with(
        PolarisClient.SupportedQueues.DEBUG
    )
    assert job_status == expected_status


def test_polaris_cluster_run_job_fails(mock_datetime, mock_polaris_client):
    cluster = PolarisCluster("debug.name", mock_polaris_client)
    mock_polaris_client.submit_job.return_value = "234"
    mock_polaris_client.list_jobs.return_value = [
        JobStatus(
            id="1234",
            name="some name",
            status="queued",
            metadata="",
            cluster="mycluster",
            done=False,
            state=JobState.PENDING,
        )
    ]
    with pytest.raises(RuntimeError):
        _ = cluster.run_job(_get_default_job("polaris"))


def test_polaris_cluster_down(mock_datetime, mock_polaris_client):
    cluster = PolarisCluster("debug-scaling.name", mock_polaris_client)
    cluster.down()
    # Nothing to assert, this method is a no-op.


def test_polaris_cluster_stop(mock_datetime, mock_polaris_client):
    cluster = PolarisCluster("debug-scaling.name", mock_polaris_client)
    cluster.stop()
    # Nothing to assert, this method is a no-op.



import os
import shutil
import subprocess
import sys
import time
from pathlib import Path
from typing import NamedTuple, Optional

import pytest

from oumi.core.configs import EvaluationConfig
from oumi.utils.torch_utils import device_cleanup
from tests import get_configs_dir
from tests.e2e import get_e2e_test_output_dir
from tests.markers import requires_gpus


class EvalTestConfig(NamedTuple):
    test_name: str
    config_path: Path
    skip: bool = False
    interactive_logs: bool = True

    use_simple_oumi_evaluate_command: bool = False
    """
    If True, the test will use the simple `oumi evaluate` command instead of the
    distributed version. This sometimes leads to lower GPU RAM usage.
    """

    model_max_length: Optional[int] = None
    batch_size: Optional[int] = None
    num_samples: Optional[int] = 20  # Limit the number of samples by default
    num_fewshot: Optional[int] = None
    enable_wandb: Optional[bool] = False  # Disable `wandb`` by default
    enable_vllm: Optional[bool] = False  # use vLLM inference engine instead of native


def get_eval_test_id_fn(val):
    assert isinstance(val, EvalTestConfig), f"{type(val)}: {val}"
    return val.test_name


def _test_eval_impl(
    test_config: EvalTestConfig,
    tmp_path: Path,
    *,
    cleanup_output_dir_on_success: bool = True,
    single_gpu: Optional[bool] = None,
):
    device_cleanup()
    if test_config.skip:
        pytest.skip(f"Skipped the test '{test_config.test_name}'!")
        return

    interactive_logs = test_config.interactive_logs

    test_tag = f"[{test_config.test_name}]"

    _START_TIME = time.perf_counter()
    output_dir = get_e2e_test_output_dir(test_config.test_name, tmp_path=tmp_path)
    output_dir.mkdir(parents=True, exist_ok=True)

    try:
        # Copy config file to output directory
        assert test_config.config_path.exists(), (
            f"{test_tag} Path doesn't exist: {test_config.config_path}"
        )
        assert test_config.config_path.is_file(), (
            f"{test_tag} Path is not a file: {test_config.config_path}"
        )

        # Verify the config is loadable
        try:
            eval_config = EvaluationConfig.from_yaml(test_config.config_path)
        except Exception as e:
            raise RuntimeError(
                f"Failed to load eval config from: {test_config.config_path}"
            ) from e

        cmd: list[str] = []
        if test_config.use_simple_oumi_evaluate_command:
            cmd.append("oumi evaluate")
        else:
            cmd.append("oumi distributed accelerate launch -m oumi evaluate")

        config_path = test_config.config_path
        # Overriding nested fields using OmegaConf's dot-list syntax is complicated,
        # or impossible. Let's just create a modified config copy instead.
        if test_config.num_samples is not None or test_config.num_fewshot is not None:
            for task in eval_config.tasks:
                if test_config.num_samples is not None:
                    task.num_samples = test_config.num_samples
                if test_config.num_fewshot is not None:
                    task.eval_kwargs["num_fewshot"] = test_config.num_fewshot

            config_path = (
                output_dir / f"MODIFIED_{test_config.config_path.name}"
            ).resolve()
            eval_config.to_yaml(config_path)

        cmd.extend(
            [
                "-c",
                str(config_path),
                "--run_name",
                test_config.test_name,
            ]
        )

        for param_name, param_value in [
            ("enable_wandb", test_config.enable_wandb),
        ]:
            if param_value is not None:
                cmd.append(f"--{param_name}={str(param_value)}")

        for param_name, param_value in [
            ("model_max_length", test_config.model_max_length),
            ("shard_for_eval", False if single_gpu else None),
        ]:
            if param_value is not None:
                cmd.append(f"--model.{param_name}={str(param_value)}")

        for param_name, param_value in [
            ("batch_size", test_config.batch_size),
        ]:
            if param_value is not None:
                cmd.append(f"--generation.{param_name}={str(param_value)}")

        env_vars = dict(os.environ)
        if "TOKENIZERS_PARALLELISM" not in env_vars:
            # Resolves the warning: "Avoid using `tokenizers` before the fork ..."
            env_vars["TOKENIZERS_PARALLELISM"] = "false"

        shell_command = " ".join(cmd)
        print(f"{test_tag} Running the command:\n{shell_command}\n")
        device_cleanup()
        result = subprocess.run(
            shell_command,
            shell=True,
            text=True,
            capture_output=(not interactive_logs),
            stdout=(sys.stdout if interactive_logs else None),
            stderr=(sys.stderr if interactive_logs else None),
            env=env_vars,
        )
        duration_sec = time.perf_counter() - _START_TIME
        if result.returncode == 0:
            print(f"{test_tag} Eval job successfully finished in {duration_sec:.2f}s!")
        else:
            print(
                f"{test_tag} Eval job failed with error code: {result.returncode} "
                f"in {duration_sec:.2f}s!"
            )
            if not interactive_logs:
                print(f"{test_tag} STDOUT:\n\n{result.stdout}\n\n")
                print(f"{test_tag} STDERR:\n\n{result.stderr}\n\n")
            assert result.returncode == 0, (
                f"{test_tag} Evaluation failed with error code: {result.returncode}"
                + ("" if interactive_logs else f"\nSTDERR:\n\n{result.stderr}\n")
            )

    except Exception as e:
        duration_sec = time.perf_counter() - _START_TIME
        print(f"{test_tag} Test failed: {str(e)}")
        print(f"{test_tag} Test duration: {duration_sec:.2f}s")
        print(f"{test_tag} Test artifacts can be found in: {output_dir}")
        raise

    if cleanup_output_dir_on_success:
        # Clean-up temp data to stay under disk quota.
        print(f"{test_tag} Cleaning up output dir on success: '{output_dir}'...")
        shutil.rmtree(output_dir)


@requires_gpus(count=1, min_gb=24.0)
@pytest.mark.parametrize(
    "test_config",
    [
        EvalTestConfig(
            test_name="eval_text_smollm_135m_single_gpu",
            config_path=(
                get_configs_dir()
                / "recipes"
                / "smollm"
                / "evaluation"
                / "135m"
                / "eval.yaml"
            ),
            num_samples=4,
        ),
        EvalTestConfig(
            test_name="eval_text_phi3_single_gpu",
            config_path=(
                get_configs_dir() / "recipes" / "phi3" / "evaluation" / "eval.yaml"
            ),
            num_samples=4,
            use_simple_oumi_evaluate_command=True,
        ),
        EvalTestConfig(
            test_name="eval_text_llama31_8b_single_gpu",
            config_path=(
                get_configs_dir()
                / "recipes"
                / "llama3_1"
                / "evaluation"
                / "8b_eval.yaml"
            ),
            num_samples=4,
        ),
        EvalTestConfig(
            test_name="eval_text_llama31_8b_vllm_single_gpu",
            config_path=(
                get_configs_dir()
                / "recipes"
                / "llama3_1"
                / "evaluation"
                / "8b_eval.yaml"
            ),
            num_samples=4,
            enable_vllm=True,
        ),
    ],
    ids=get_eval_test_id_fn,
)
@pytest.mark.e2e
@pytest.mark.single_gpu
def test_eval_text_1gpu_24gb(test_config: EvalTestConfig, tmp_path: Path):
    _test_eval_impl(
        test_config=test_config,
        tmp_path=tmp_path,
        single_gpu=True,
    )


@requires_gpus(count=1, min_gb=24.0)
@pytest.mark.parametrize(
    "test_config",
    [
        EvalTestConfig(
            test_name="eval_mm_llama32v_11b_single_gpu",
            config_path=(
                get_configs_dir()
                / "recipes"
                / "vision"
                / "llama3_2_vision"
                / "evaluation"
                / "11b_eval.yaml"
            ),
            num_samples=4,
            num_fewshot=2,
        ),
    ],
    ids=get_eval_test_id_fn,
)
@pytest.mark.e2e
@pytest.mark.single_gpu
def test_eval_multimodal_1gpu_24gb(test_config: EvalTestConfig, tmp_path: Path):
    _test_eval_impl(
        test_config=test_config,
        tmp_path=tmp_path,
        single_gpu=True,
    )


@requires_gpus(count=4, min_gb=39.0)
@pytest.mark.parametrize(
    "test_config",
    [
        EvalTestConfig(
            test_name="eval_text_llama31_70b_multi_gpu",
            config_path=(
                get_configs_dir()
                / "recipes"
                / "llama3_1"
                / "evaluation"
                / "70b_eval.yaml"
            ),
            num_samples=4,
            num_fewshot=2,
            use_simple_oumi_evaluate_command=True,
        ),
    ],
    ids=get_eval_test_id_fn,
)
@pytest.mark.e2e
@pytest.mark.multi_gpu
def test_eval_text_4gpu_40gb(test_config: EvalTestConfig, tmp_path: Path):
    _test_eval_impl(
        test_config=test_config,
        tmp_path=tmp_path,
    )


@requires_gpus(count=4, min_gb=24.0)
@pytest.mark.parametrize(
    "test_config",
    [
        EvalTestConfig(
            test_name="eval_mm_llama32v_11b_multi_gpu",
            config_path=(
                get_configs_dir()
                / "recipes"
                / "vision"
                / "llama3_2_vision"
                / "evaluation"
                / "11b_eval.yaml"
            ),
            num_samples=4,
            num_fewshot=2,
            use_simple_oumi_evaluate_command=True,
        ),
    ],
    ids=get_eval_test_id_fn,
)
@pytest.mark.e2e
@pytest.mark.multi_gpu
def test_eval_multimodal_4gpu_24gb(test_config: EvalTestConfig, tmp_path: Path):
    _test_eval_impl(
        test_config=test_config,
        tmp_path=tmp_path,
    )



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import Any, Optional

from oumi.core.configs import TrainingConfig


class BaseTrainer(ABC):
    @abstractmethod
    def train(self, resume_from_checkpoint: Optional[str] = None) -> None:
        """Trains a model."""

    @abstractmethod
    def save_state(self) -> None:
        """Saves the Trainer state.

        Under distributed environment this is done only for a process with rank 0.
        """
        # TODO: Define semantics of this method more clearly.
        # Can it be merged with save_model()?

    @abstractmethod
    def save_model(self, config: TrainingConfig, final: bool = True) -> None:
        """Saves the model's state dictionary to the specified output directory.

        Args:
            config (TrainingConfig): The Oumi training config.
            final (bool): Whether this is the final model being saved during training.

        Returns:
            None
        """
        # TODO: Define semantics of this method more clearly.

    @abstractmethod
    def get_last_eval_metrics(self) -> dict[str, Any]:
        """Returns the last evaluation metrics from the trained model.

        Returns:
            dict with the evaluation metrics from the last trained model step.
        """



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oumi.core.datasets import VisionLanguageDpoDataset
from oumi.core.registry import register_dataset


@register_dataset("openbmb/RLAIF-V-Dataset")
class OpenbmbRlaifVDataset(VisionLanguageDpoDataset):
    """Preprocess the RLAIF-V dataset for DPO.

    See Also:
        For more information on how to use this dataset, refer to:
        - Huggingface hub: https://huggingface.co/datasets/openbmb/RLAIF-V-Dataset
    """

    default_dataset = "openbmb/RLAIF-V-Dataset"

    def __init__(self, *args, **kwargs):
        """Initialize the OpenbmbRlaifVDataset."""
        super().__init__(
            *args,
            images_key="image",
            prompt_key="question",
            chosen_key="chosen",
            rejected_key="rejected",
            **kwargs,
        )



import torch
import torch.nn as nn
from einops import rearrange

# from timm.layers import LayerNorm, LayerNorm2d
# from timm.models.regnet import RegStage


def build_pos_embeds(num_input_tokens: int, vision_hidden_size: int):
    # pos emb
    pos_emb = torch.nn.Parameter(torch.zeros(1, num_input_tokens, vision_hidden_size))
    nn.init.trunc_normal_(pos_emb, mean=0.0, std=0.02)

    return pos_emb


def build_mlp(depth, hidden_size, output_hidden_size):
    layers = [nn.Linear(hidden_size, output_hidden_size)]
    for _ in range(1, depth):
        layers.append(nn.SiLU())
        layers.append(nn.Linear(output_hidden_size, output_hidden_size))
    return nn.Sequential(*layers)


class Projector(nn.Module):
    """Base projector class"""

    def __init__(
        self,
        encoder_hidden_size: int,
        num_input_tokens: int,
        num_queries: int,
        output_hidden_size: int,
    ):
        super().__init__()
        self.encoder_hidden_size = encoder_hidden_size
        self.num_input_tokens = num_input_tokens
        self.num_queries = num_queries
        self.hidden_size = 1024
        self.output_hidden_size = output_hidden_size

        # pos emb
        self.pos_emb = build_pos_embeds(num_input_tokens, encoder_hidden_size)
        # self.pos_emb = None

        self.build_net()

    def build_net(self):
        raise NotImplementedError

    def _forward(self, x):
        raise NotImplementedError

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Args:
        x: (B, L, encoder_hidden_size) tensor from the visual backbone (CLIP visual encoder), including cls token.
        """
        if self.pos_emb is not None:
            x = x + self.pos_emb

        dtype = x.dtype
        # x = self._forward(x.to(torch.float32))  # (B, L, output_hidden_size)
        x = self._forward(x)

        return x.to(dtype)


class ConvProjector(Projector):
    def _forward(self, x):
        hw = int(x.size(1) ** 0.5)
        x = rearrange(x, "b (h w) d -> b d h w", h=hw, w=hw)
        x = self.net(x)
        x = rearrange(x, "b d h w -> b (h w) d")
        x = self.readout(x)
        return x


class CAbstractor(nn.Module):
    """Base projector class"""

    def __init__(
        self,
        encoder_hidden_size: int,
        output_hidden_size: int,
    ):
        super().__init__()
        self.encoder_hidden_size = encoder_hidden_size
        self.hidden_size = 1024
        self.output_hidden_size = output_hidden_size

        # pos emb
        # self.pos_emb = build_pos_embeds(num_input_tokens, encoder_hidden_size)
        self.pos_emb = None

        self.downsamples = nn.Conv2d(
            encoder_hidden_size,
            self.hidden_size,
            kernel_size=3,
            stride=2,
            padding=1,
            bias=False,
        )

        self.readout = nn.Sequential(
            nn.Linear(self.hidden_size, output_hidden_size),
            nn.GELU(),
            nn.Linear(output_hidden_size, output_hidden_size),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        dtype = x.dtype
        x = x.to(torch.float32)

        hw = int(x.size(1) ** 0.5)
        x = rearrange(x, "b (h w) d -> b d h w", h=hw, w=hw)
        x = self.downsamples(x)
        x = rearrange(x, "b d h w -> b (h w) d")
        x = self.readout(x)

        return x.to(dtype)


# class CAbstractor(ConvProjector):
#     """C-Abstractor"""
#     def build_net(self):
#         encoder_hidden_size = self.encoder_hidden_size
#         hidden_size = self.hidden_size
#         output_hidden_size = self.output_hidden_size
#         depth = 3
#         mlp_depth = 2

#         n_queries = self.num_queries
#         assert (n_queries ** 0.5).is_integer(), "n_queries must be square number"
#         hw = int(n_queries ** 0.5)

#         # RegBlock = ResBlock + SE
#         RegBlock = partial(
#             RegStage,
#             stride=1,
#             dilation=1,
#             act_layer=nn.SiLU,
#             norm_layer=LayerNorm2d,
#         )

#         s1 = RegBlock(
#             depth,
#             encoder_hidden_size,
#             hidden_size,
#         )
#         sampler = nn.AdaptiveAvgPool2d((hw, hw))
#         s2 = RegBlock(
#             depth,
#             hidden_size,
#             hidden_size,
#         )

#         self.net = nn.Sequential(s1, sampler, s2)
#         self.readout = build_mlp(mlp_depth, hidden_size, output_hidden_size)



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

from transformers import SpecialTokensMixin

from oumi.core.tokenizers import BaseTokenizer
from oumi.utils.logging import logger

# Llama 3.1/3.2 models already have `<|finetune_right_pad_id|>` token in their vocab.
LLAMA_SPECIAL_TOKENS_MIXIN = SpecialTokensMixin(pad_token="<|finetune_right_pad_id|>")

special_tokens = {
    "meta-llama/Llama-3.1-8B": LLAMA_SPECIAL_TOKENS_MIXIN,
    "meta-llama/Llama-3.1-8B-Instruct": LLAMA_SPECIAL_TOKENS_MIXIN,
    "meta-llama/Llama-3.1-70B": LLAMA_SPECIAL_TOKENS_MIXIN,
    "meta-llama/Llama-3.1-70B-Instruct": LLAMA_SPECIAL_TOKENS_MIXIN,
    "meta-llama/Llama-3.1-405B": LLAMA_SPECIAL_TOKENS_MIXIN,
    "meta-llama/Llama-3.1-405B-Instruct": LLAMA_SPECIAL_TOKENS_MIXIN,
    "meta-llama/Llama-3.1-405B-FP8": LLAMA_SPECIAL_TOKENS_MIXIN,
    "meta-llama/Llama-3.1-405B-Instruct-FP8": LLAMA_SPECIAL_TOKENS_MIXIN,
    "meta-llama/Meta-Llama-3.1-8B": LLAMA_SPECIAL_TOKENS_MIXIN,
    "meta-llama/Meta-Llama-3.1-8B-Instruct": LLAMA_SPECIAL_TOKENS_MIXIN,
    "meta-llama/Meta-Llama-3.1-70B": LLAMA_SPECIAL_TOKENS_MIXIN,
    "meta-llama/Meta-Llama-3.1-70B-Instruct": LLAMA_SPECIAL_TOKENS_MIXIN,
    "meta-llama/Meta-Llama-3.1-405B": LLAMA_SPECIAL_TOKENS_MIXIN,
    "meta-llama/Meta-Llama-3.1-405B-Instruct": LLAMA_SPECIAL_TOKENS_MIXIN,
    "meta-llama/Meta-Llama-3.1-405B-FP8": LLAMA_SPECIAL_TOKENS_MIXIN,
    "meta-llama/Meta-Llama-3.1-405B-Instruct-FP8": LLAMA_SPECIAL_TOKENS_MIXIN,
    "meta-llama/Llama-3.2-1B": LLAMA_SPECIAL_TOKENS_MIXIN,
    "meta-llama/Llama-3.2-1B-Instruct": LLAMA_SPECIAL_TOKENS_MIXIN,
    "meta-llama/Llama-3.2-3B": LLAMA_SPECIAL_TOKENS_MIXIN,
    "meta-llama/Llama-3.2-3B-Instruct": LLAMA_SPECIAL_TOKENS_MIXIN,
}

# Lowercase all keys for case-insensitive lookup.
special_tokens = {k.lower(): v for k, v in special_tokens.items()}


def get_default_special_tokens(
    tokenizer: Optional[BaseTokenizer],
) -> SpecialTokensMixin:
    """Returns the default special tokens for the tokenizer that was provided.

    Args:
        tokenizer: The tokenizer to get special tokens for.

    Returns:
        The special tokens mixin for the tokenizer.

    Description:
        This function looks up the special tokens for the provided tokenizer, for a list
        of known models. If the tokenizer is not recognized, it returns an empty special
        tokens mixin. This function is used as a fallback mechanism when a special token
        is required, but is not provided in the tokenizer's configuration. The primary
        use case for this is to retrieve the padding special token (`pad_token`), which
        is oftentimes not included in the tokenizer's configuration, even if it exists
        in the tokenizer's vocabulary.
    """
    if tokenizer and tokenizer.name_or_path:
        if tokenizer.name_or_path.lower() in special_tokens:
            return special_tokens[tokenizer.name_or_path.lower()]
        else:
            logger.warning(
                f"Special tokens lookup for tokenizer {tokenizer.name_or_path} failed."
            )
    return SpecialTokensMixin()



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from enum import Enum
from typing import Optional

import torch.distributed.fsdp as torch_fsdp

from oumi.core.configs.params.base_params import BaseParams


class ShardingStrategy(str, Enum):
    """The sharding strategies for FullyShardedDataParallel (FSDP).

    See :external:class:`torch.distributed.fsdp.ShardingStrategy`
    for more details.
    """

    FULL_SHARD = "FULL_SHARD"
    """Shards model parameters, gradients, and optimizer states.
    Provides the most memory efficiency but may impact performance."""

    SHARD_GRAD_OP = "SHARD_GRAD_OP"
    """Shards gradients and optimizer states, but not model
    parameters. Balances memory savings and performance."""

    HYBRID_SHARD = "HYBRID_SHARD"
    """Shards model parameters within a node and replicates them
    across nodes."""

    NO_SHARD = "NO_SHARD"
    """No sharding is applied. Parameters, gradients, and optimizer states
    are kept in full on each GPU."""

    HYBRID_SHARD_ZERO2 = "HYBRID_SHARD_ZERO2"
    """Apply SHARD_GRAD_OP within a node, and replicate
    parameters across nodes."""

    def to_torch(self) -> torch_fsdp.ShardingStrategy:
        """Convert the enum to the corresponding torch_fsdp.ShardingStrategy."""
        strategy_map = {
            ShardingStrategy.FULL_SHARD: torch_fsdp.ShardingStrategy.FULL_SHARD,
            ShardingStrategy.SHARD_GRAD_OP: torch_fsdp.ShardingStrategy.SHARD_GRAD_OP,
            ShardingStrategy.HYBRID_SHARD: torch_fsdp.ShardingStrategy.HYBRID_SHARD,
            ShardingStrategy.NO_SHARD: torch_fsdp.ShardingStrategy.NO_SHARD,
            ShardingStrategy.HYBRID_SHARD_ZERO2: (
                torch_fsdp.ShardingStrategy._HYBRID_SHARD_ZERO2
            ),
        }

        if self not in strategy_map:
            raise ValueError(f"Unsupported sharding strategy: {self}")

        return strategy_map[self]


class StateDictType(str, Enum):
    """The supported state dict types for FullyShardedDataParallel (FSDP).

    This controls how the model's state dict will be saved during checkpointing, and
    how it can be consumed afterwards.
    """

    FULL_STATE_DICT = "FULL_STATE_DICT"
    """The state dict will be saved in a non-sharded, unflattened format.

    This is similar to checkpointing without FSDP.
    """

    SHARDED_STATE_DICT = "SHARDED_STATE_DICT"
    """The state dict will be saved in a sharded, unflattened format.

    This can be used by other parallel schemes.
    """

    LOCAL_STATE_DICT = "LOCAL_STATE_DICT"
    """The state dict will be saved in a sharded, flattened format.

    Since it's flattened, this can only be used by FSDP.
    """

    def to_torch(self) -> torch_fsdp.StateDictType:
        """Converts to the corresponding torch.distributed.fsdp.StateDictType."""
        state_dict_map = {
            StateDictType.FULL_STATE_DICT: torch_fsdp.StateDictType.FULL_STATE_DICT,
            StateDictType.SHARDED_STATE_DICT: (
                torch_fsdp.StateDictType.SHARDED_STATE_DICT
            ),
            StateDictType.LOCAL_STATE_DICT: torch_fsdp.StateDictType.LOCAL_STATE_DICT,
        }

        if self not in state_dict_map:
            raise ValueError(f"Unsupported state dict type: {self}")

        return state_dict_map[self]


class BackwardPrefetch(str, Enum):
    """The backward prefetch options for FullyShardedDataParallel (FSDP)."""

    BACKWARD_PRE = "BACKWARD_PRE"
    """Enables the most overlap but increases memory usage the most."""

    BACKWARD_POST = "BACKWARD_POST"
    """Enables less overlap but requires less memory usage."""

    NO_PREFETCH = "NO_PREFETCH"
    """Disables backward prefetching altogether."""

    def to_torch(self) -> Optional[torch_fsdp.BackwardPrefetch]:
        """Convert the enum to the corresponding torch_fsdp.BackwardPrefetch."""
        map = {
            BackwardPrefetch.BACKWARD_PRE: torch_fsdp.BackwardPrefetch.BACKWARD_PRE,
            BackwardPrefetch.BACKWARD_POST: torch_fsdp.BackwardPrefetch.BACKWARD_POST,
            BackwardPrefetch.NO_PREFETCH: None,
        }

        if self not in map:
            raise ValueError(f"Unsupported backward prefetch option: {self}")
        return map[self]


class AutoWrapPolicy(str, Enum):
    """The auto wrap policies for FullyShardedDataParallel (FSDP)."""

    SIZE_BASED_WRAP = "SIZE_BASED_WRAP"
    """Wraps layers based on parameter count."""

    TRANSFORMER_BASED_WRAP = "TRANSFORMER_BASED_WRAP"
    """Wraps layers based on the transformer block layer."""

    NO_WRAP = "NO_WRAP"
    """No automatic wrapping is performed."""


@dataclass
class FSDPParams(BaseParams):
    """Configuration options for Pytorch's FullyShardedDataParallel (FSDP) training."""

    enable_fsdp: bool = False
    """If True, enables FullyShardedDataParallel training.

    Allows training larger models by sharding models and gradients across multiple GPUs.
    """

    sharding_strategy: ShardingStrategy = ShardingStrategy.FULL_SHARD
    """Determines how to shard model parameters across GPUs.

    See :external:class:`torch.distributed.fsdp.api.ShardingStrategy` for more details.


    Options:
        FULL_SHARD: Shards model parameters, gradients, and optimizer states.
            Provides the most memory efficiency but may impact performance.
        SHARD_GRAD_OP: Shards gradients and optimizer states, but not model
            parameters. Balances memory savings and performance.
        HYBRID_SHARD: Shards model parameters within a node and replicates them
            across nodes.
        NO_SHARD: No sharding is applied. Parameters, gradients, and optimizer states
            are kept in full on each GPU.
        HYBRID_SHARD_ZERO2: Apply SHARD_GRAD_OP within a node, and replicate
            parameters across nodes.

    Warning:
        NO_SHARD option is deprecated and will be removed in a future release.
            Please use DistributedDataParallel (DDP) instead.
    """

    cpu_offload: bool = False
    """If True, offloads parameters and gradients to CPU when not in use."""

    mixed_precision: Optional[str] = None
    """Enables mixed precision training.

    Options: None, "fp16", "bf16".
    """

    backward_prefetch: BackwardPrefetch = BackwardPrefetch.BACKWARD_PRE
    """Determines when to prefetch the next set of parameters.

    Improves throughput by enabling communication and computation overlap
    in the backward pass at the cost of slightly increased memory usage.

    Options:
        BACKWARD_PRE: Enables the most overlap but increases memory
            usage the most. This prefetches the next set of parameters *before*
            the current set of parameters' gradient computation.
        BACKWARD_POST: Enables less overlap but requires less memory
            usage. This prefetches the next set of parameters *after* the current
            set of parameters' gradient computation.
        NO_PREFETCH: Disables backward prefetching altogether. This has no overlap and
            does not increase memory usage. This may degrade throughput significantly.
    """

    forward_prefetch: bool = False
    """If True, prefetches the forward pass results."""

    use_orig_params: Optional[bool] = None
    """If True, uses the PyTorch Module's original parameters for FSDP.

    For more information, see: https://pytorch.org/docs/stable/fsdp.html.
    If not specified, it will be automatically inferred based on other config values.
    """

    state_dict_type: StateDictType = StateDictType.FULL_STATE_DICT
    """Specifies the type of state dict to use for checkpointing."""

    auto_wrap_policy: AutoWrapPolicy = AutoWrapPolicy.NO_WRAP
    """Policy for automatically wrapping layers in FSDP."""

    min_num_params: int = 100_000
    """Minimum number of parameters for a layer to be wrapped when using
    size_based policy. This has no effect when using
    transformer_based policy.
    """

    transformer_layer_cls: Optional[str] = None
    """Class name for transformer layers when using transformer_based policy.

    This has no effect when using size_based policy.
    """

    sync_module_states: bool = True
    """If True, synchronizes module states across processes.

    When enabled, each FSDP module broadcasts parameters and buffers from rank 0
    to ensure replication across ranks.
    """



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oumi.core.datasets import BasePretrainingDataset
from oumi.core.registry import register_dataset


@register_dataset("roneneldan/TinyStories")
class TinyStoriesDataset(BasePretrainingDataset):
    """TinyStoriesDataset class for loading and processing the TinyStories dataset.

    This dataset contains synthetically generated short stories with a small
    vocabulary, created by GPT-3.5 and GPT-4. It is designed for text generation
    tasks and is available in English.

    See Also:
        - Paper: https://arxiv.org/abs/2305.07759
        - Huggingface hub: https://huggingface.co/datasets/roneneldan/TinyStories

    Note:
        The dataset is available under the CDLA-Sharing-1.0 license.
    """

    default_dataset = "roneneldan/TinyStories"



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TypeVar

T = TypeVar("T")


def batch(dataset: list[T], batch_size: int) -> list[list[T]]:
    """Batches the provided dataset.

    Args:
        dataset: The dataset to batch, which is a flat list of items.
        batch_size: The desired size of each batch.

    Returns:
        A list of batches. Each batch is a list of `batch_size` items, assuming that
        the dataset's size is a multiple of `batch_size`. Otherwise, the last batch to
        be included will contain less items than `batch_size`.
    """
    batches = []
    for dataset_index in range(0, len(dataset), batch_size):
        batches.append(dataset[dataset_index : dataset_index + batch_size])
    return batches


def unbatch(dataset: list[list[T]]) -> list[T]:
    """Unbatches (flatten) the provided dataset."""
    return [item for batch in dataset for item in batch]



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Unit tests for quantizer builders."""

from unittest.mock import Mock, patch

import pytest  # type: ignore

from oumi.builders.quantizers import (
    build_quantizer,
    get_available_methods,
    get_supported_formats,
    list_all_methods,
)
from oumi.quantize.base import BaseQuantization


class TestQuantizerBuilders:
    """Test cases for quantizer builder functions."""

    def test_build_quantizer_awq_method(self):
        """Test building AWQ quantizer for AWQ methods."""
        # Test various AWQ methods
        for method in ["awq_q4_0", "awq_q4_1", "awq_q8_0", "awq_f16"]:
            quantizer = build_quantizer(method)

            # Verify correct type
            assert quantizer.__class__.__name__ == "AwqQuantization"
            assert isinstance(quantizer, BaseQuantization)

    def test_build_quantizer_bnb_method(self):
        """Test building BnB quantizer for BnB methods."""
        # Test BnB methods
        for method in ["bnb_4bit", "bnb_8bit"]:
            quantizer = build_quantizer(method)

            # Verify correct type
            assert quantizer.__class__.__name__ == "BitsAndBytesQuantization"
            assert isinstance(quantizer, BaseQuantization)

    def test_build_quantizer_unsupported_method(self):
        """Test building quantizer with unsupported method."""
        with pytest.raises(
            ValueError, match="Unsupported quantization method: invalid_method"
        ):
            build_quantizer("invalid_method")

    @patch("oumi.quantize.awq_quantizer.AwqQuantization")
    @patch("oumi.quantize.bnb_quantizer.BitsAndBytesQuantization")
    def test_build_quantizer_fallback(self, mock_bnb, mock_awq):
        """Test build_quantizer fallback when method doesn't match prefix."""
        # Setup mocks
        mock_awq_instance = Mock()
        mock_awq_instance.supports_method.return_value = False
        mock_awq.return_value = mock_awq_instance

        mock_bnb_instance = Mock()
        mock_bnb_instance.supports_method.return_value = True
        mock_bnb.return_value = mock_bnb_instance

        # Test with method that doesn't match any prefix
        result = build_quantizer("custom_method")

        # Verify BNB instance was returned (since it supports the method)
        assert result == mock_bnb_instance

        # Verify both were checked
        mock_awq_instance.supports_method.assert_called_with("custom_method")
        mock_bnb_instance.supports_method.assert_called_with("custom_method")

    def test_get_available_methods(self):
        """Test getting all available quantization methods."""
        methods = get_available_methods()

        # Verify structure
        assert isinstance(methods, dict)
        assert "AWQ" in methods
        assert "BitsAndBytes" in methods

        # Verify AWQ methods
        assert "awq_q4_0" in methods["AWQ"]
        assert "awq_q4_1" in methods["AWQ"]
        assert "awq_q8_0" in methods["AWQ"]
        assert "awq_f16" in methods["AWQ"]

        # Verify BitsAndBytes methods
        assert "bnb_4bit" in methods["BitsAndBytes"]
        assert "bnb_8bit" in methods["BitsAndBytes"]

    def test_get_supported_formats(self):
        """Test getting all supported output formats."""
        formats = get_supported_formats()

        # Verify it's a list
        assert isinstance(formats, list)

        # Verify expected formats
        assert "safetensors" in formats

        # Verify sorted
        assert formats == sorted(formats)

    def test_list_all_methods(self):
        """Test listing all available methods."""
        all_methods = list_all_methods()

        # Verify it's a list
        assert isinstance(all_methods, list)

        # Verify contains expected methods
        expected_methods = [
            "awq_q4_0",
            "awq_q4_1",
            "awq_q8_0",
            "awq_f16",
            "bnb_4bit",
            "bnb_8bit",
        ]
        for method in expected_methods:
            assert method in all_methods

        # Verify sorted
        assert all_methods == sorted(all_methods)

    def test_get_available_methods_structure(self):
        """Test the structure of available methods dictionary."""
        methods = get_available_methods()

        # Each value should be a list
        for quantizer_name, method_list in methods.items():
            assert isinstance(method_list, list)
            assert len(method_list) > 0

            # Each method should be a string
            for method in method_list:
                assert isinstance(method, str)

    @patch("oumi.quantize.awq_quantizer.AwqQuantization")
    def test_build_quantizer_creates_new_instance(self, mock_awq):
        """Test that build_quantizer creates new instances each time."""
        # Setup mock
        mock_awq.side_effect = [Mock(), Mock()]  # Return different instances

        # Build two quantizers
        q1 = build_quantizer("awq_q4_0")
        q2 = build_quantizer("awq_q4_0")

        # Verify they are different instances
        assert q1 is not q2

        # Verify constructor was called twice
        assert mock_awq.call_count == 2



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import os
from pathlib import Path
from typing import Optional, Union

import transformers

from oumi.core.configs.internal.supported_models import (
    is_custom_model,
)
from oumi.utils.logging import logger


def is_cached_to_disk_hf_dataset(dataset_folder: Union[str, Path]) -> bool:
    """Detects whether a dataset was saved using `dataset.save_to_disk()`.

    Such datasets should be loaded using `datasets.Dataset.load_from_disk()`

    Returns:
        Whether the dataset was saved using `dataset.save_to_disk()` method.
    """
    if not dataset_folder:
        return False

    dataset_path: Path = Path(dataset_folder)

    if dataset_path.exists() and dataset_path.is_dir():
        for file_name in ("dataset_info.json", "state.json"):
            file_path: Path = dataset_path / file_name
            if not (file_path.exists() and file_path.is_file()):
                logger.warning(
                    f"The dataset {str(dataset_path)} is missing "
                    f"a required file: {file_name}."
                )
                return False
        return True

    return False


def find_hf_token() -> Optional[str]:
    """Attempts to find HuggingFace access token.

    Returns:
        A valid HF access token, or `None` if not found.
    """
    hf_token = os.environ.get("HF_TOKEN", None)
    if not hf_token:
        _DEFAULT_HF_HOME_PATH = "~/.cache/huggingface"

        file_must_exist = False
        token_path = os.environ.get("HF_TOKEN_PATH", None)
        if token_path:
            token_path = Path(token_path)
            file_must_exist = True
        else:
            hf_home_dir = Path(
                os.environ.get("HF_HOME", _DEFAULT_HF_HOME_PATH)
                or _DEFAULT_HF_HOME_PATH
            )
            token_path = hf_home_dir / "token"

        if token_path.exists():
            if token_path.is_file():
                hf_token = token_path.read_text().strip()
        elif file_must_exist:
            raise FileNotFoundError(f"Missing HF token file: '{token_path}'")

    return hf_token if hf_token else None


@functools.cache
def get_hf_chat_template(
    tokenizer_name: str, *, trust_remote_code: bool = False
) -> Optional[str]:
    """Returns chat template provided by HF for `tokenizer_name`."""
    if not tokenizer_name or is_custom_model(tokenizer_name):
        return None

    tokenizer = transformers.AutoTokenizer.from_pretrained(
        tokenizer_name, trust_remote_code=trust_remote_code
    )
    if tokenizer.chat_template:
        if not isinstance(tokenizer.chat_template, str):
            raise RuntimeError(
                f"Chat template for tokenizer_name: {tokenizer_name} "
                "is not a string! "
                f"Actual type: {type(tokenizer.chat_template)}"
            )
        return tokenizer.chat_template
    return None



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Based on MFU from PaLM paper: https://arxiv.org/pdf/2204.02311."""

from typing import Optional

import torch

_TFLOPS = "tflops"
_DEVICE_SPECS = {
    # https://www.nvidia.com/content/dam/en-zz/Solutions/Data-Center/a100/pdf/nvidia-a100-datasheet-us-nvidia-1758950-r4-web.pdf
    "NVIDIA A100-PCIE-40GB": {
        _TFLOPS: {
            torch.float32: 19.5,
            torch.float16: 312.0,
            torch.bfloat16: 312.0,
        },
    },
    "NVIDIA A100-PCIE-80GB": {
        _TFLOPS: {
            torch.float32: 19.5,
            torch.float16: 312.0,
            torch.bfloat16: 312.0,
        },
    },
    "NVIDIA A100 80GB PCIe": {
        _TFLOPS: {
            torch.float32: 19.5,
            torch.float16: 312.0,
            torch.bfloat16: 312.0,
        },
    },
    "NVIDIA A100-SXM4-40GB": {
        _TFLOPS: {
            torch.float32: 19.5,
            torch.float16: 312.0,
            torch.bfloat16: 312.0,
        }
    },
    "NVIDIA A100-SXM4-80GB": {
        _TFLOPS: {
            torch.float32: 19.5,
            torch.float16: 312.0,
            torch.bfloat16: 312.0,
        }
    },
    # https://www.nvidia.com/content/PDF/nvidia-ampere-ga-102-gpu-architecture-whitepaper-v2.1.pdf
    "NVIDIA GeForce RTX 3090": {
        _TFLOPS: {
            torch.float32: 35.6,
            torch.float16: 71.0,
            torch.bfloat16: 71.0,
        },
    },
    # Only used for testing purposes
    # https://cloud.google.com/tpu/docs/v4
    "TPUv4": {
        _TFLOPS: {
            torch.float16: 275.0,
            torch.bfloat16: 275.0,
        },
    },
    # https://www.nvidia.com/en-us/data-center/l4/
    # Note that values in that page are shown with sparsity.
    "NVIDIA L4": {
        _TFLOPS: {
            torch.float32: 60.0,
            torch.float16: 121.0,
            torch.bfloat16: 121.0,
        },
    },
    # https://www.nvidia.com/en-us/data-center/tesla-t4/
    "Tesla T4": {
        _TFLOPS: {
            torch.float32: 8.1,
            torch.float16: 65.0,
            torch.bfloat16: 65.0,
        },
    },
}


def _get_device_flops(device_name: str, dtype: torch.dtype) -> float:
    """Returns peak TFLOPS for the given device name and dtype."""
    if device_name not in _DEVICE_SPECS:
        raise NotImplementedError(
            f"Unknown device name for getting hardware flops: {device_name}"
        )

    specs = _DEVICE_SPECS[device_name]
    if dtype not in specs[_TFLOPS]:
        raise NotImplementedError(f"Unknown dtype {dtype} for device {device_name}")

    return specs[_TFLOPS][dtype] * 1e12


def _get_model_flops_per_token(
    num_params: int,
    num_layers: Optional[int] = None,
    num_attention_heads: Optional[int] = None,
    attention_head_size: Optional[int] = None,
    sequence_length: Optional[int] = None,
    add_rematerialization: bool = False,
) -> int:
    """Returns the number of FLOPs per token for the given model configuration."""
    if num_params <= 0:
        raise ValueError(f"Must have a positive number of model params: {num_params}")

    forward_flops = 2 * num_params
    backward_flops = 4 * num_params
    attention_flops = 0
    if num_layers and num_attention_heads and attention_head_size and sequence_length:
        attention_flops = (
            sequence_length
            * num_layers
            * num_attention_heads
            * attention_head_size
            * 12
        )

    rematerialization_flops = 0
    if add_rematerialization:
        # FIXME: Needs to be calculated based on checkpointing configuration
        # 73% of forward and all of attention
        # PaLM paper mentions 75%, but the calculated value requires 73%, paper error?
        rematerialization_flops = int(0.73 * forward_flops + attention_flops)

    return forward_flops + backward_flops + attention_flops + rematerialization_flops


def calculate_mfu_from_model_flops_per_second(
    device_name: str,
    num_devices: int,
    dtype: torch.dtype,
    model_flops_per_second_on_all_devices: float,
) -> float:
    """Returns the number of MFU for the given model flops per second."""
    if num_devices <= 0:
        raise ValueError(f"Must have a positive number of devices: {num_devices}")
    device_flops_per_second = _get_device_flops(device_name, dtype) * num_devices
    model_flop_utilization = (
        model_flops_per_second_on_all_devices / device_flops_per_second
    )
    return model_flop_utilization


def calculate_mfu(
    device_name: str,
    num_devices: int,
    dtype: torch.dtype,
    num_params: int,
    num_tokens: int,
    delta_time_seconds: float,
    num_layers: Optional[int] = None,
    num_attention_heads: Optional[int] = None,
    attention_head_size: Optional[int] = None,
    sequence_length: Optional[int] = None,
    add_rematerialization: bool = False,
) -> float:
    """Returns the number of MFU for the given model configuration."""
    if num_tokens <= 0:
        raise ValueError(f"Must have a positive number of tokens: {num_tokens}")
    if delta_time_seconds <= 0:
        raise ValueError(f"Must have a positive delta time: {delta_time_seconds}")

    model_flops_per_token = _get_model_flops_per_token(
        num_params,
        num_layers,
        num_attention_heads,
        attention_head_size,
        sequence_length,
        add_rematerialization,
    )
    tokens_per_second = num_tokens / delta_time_seconds
    model_flops_per_second = model_flops_per_token * tokens_per_second

    return calculate_mfu_from_model_flops_per_second(
        device_name=device_name,
        num_devices=num_devices,
        dtype=dtype,
        model_flops_per_second_on_all_devices=model_flops_per_second,
    )



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oumi.core.registry import register_dataset
from oumi.datasets.sft.alpaca import AlpacaDataset


@register_dataset("uiuc-convai/CoALM-IT")
class CoALMDataset(AlpacaDataset):
    """Dataset class for the UIUC CoALM dataset.

    This dataset follows the same structure as the Alpaca dataset, with
    instruction, input, and output fields. It is designed for training
    Conversational Agentic Language Models (CoALM) that can handle both
    task-oriented dialogue and function calling.


    Dataset Sources:
        - Paper: https://arxiv.org/abs/2502.08820
        - Project Page: https://emrecanacikgoz.github.io/CoALM/
        - Repository: https://github.com/oumi-ai/oumi/tree/main/configs/projects/calm
        - Dataset: https://huggingface.co/datasets/uiuc-convai/CoALM-IT

    Examples:
        >>> from oumi.datasets import CoALMDataset
        >>> dataset = CoALMDataset()
        >>> # The dataset will be loaded from HuggingFace with the path
        >>> # "uiuc-convai/CoALM-IT" and transformed into the Oumi
        >>> # conversation format automatically.
    """

    default_dataset = "uiuc-convai/CoALM-IT"



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Builders module for the Oumi (Open Universal Machine Intelligence) library.

This module provides builder functions to construct and configure
different components of the Oumi framework, including datasets, models,
optimizers, and trainers.

The builder functions encapsulate the complexity of creating these components,
allowing for easier setup and configuration of machine learning experiments.
"""

from oumi.builders.callbacks import build_training_callbacks
from oumi.builders.collators import build_collator_from_config, build_data_collator
from oumi.builders.data import (
    build_dataset,
    build_dataset_mixture,
)
from oumi.builders.metrics import build_metrics_function
from oumi.builders.models import (
    build_chat_template,
    build_model,
    build_peft_model,
    build_tokenizer,
    is_image_text_llm,
)
from oumi.builders.optimizers import build_optimizer
from oumi.builders.processors import build_processor
from oumi.builders.quantizers import build_quantizer
from oumi.builders.rewards import build_reward_functions
from oumi.builders.rollouts import build_rollout_function
from oumi.builders.training import build_trainer
from oumi.builders.tuning import build_tuner

__all__ = [
    "build_chat_template",
    "build_dataset_mixture",
    "build_dataset",
    "build_data_collator",
    "build_metrics_function",
    "build_model",
    "build_optimizer",
    "build_peft_model",
    "build_processor",
    "build_quantizer",
    "build_reward_functions",
    "build_rollout_function",
    "build_tokenizer",
    "build_trainer",
    "build_tuner",
    "build_training_callbacks",
    "is_image_text_llm",
    "build_collator_from_config",
]



import pytest

from oumi.core.configs.params.evaluation_params import (
    AlpacaEvalTaskParams,
    EvaluationBackend,
    EvaluationTaskParams,
    LMHarnessTaskParams,
)


@pytest.mark.parametrize(
    (
        "evaluation_backend,"
        "task_name,"
        "num_samples,"
        "eval_kwargs,"
        "expected_backend,"
        "backend_class,"
    ),
    [
        # Alpaca Eval run with no arguments.
        (
            "alpaca_eval",
            "",
            None,
            {},
            EvaluationBackend.ALPACA_EVAL,
            AlpacaEvalTaskParams,
        ),
        # Alpaca Eval run with arguments.
        (
            "alpaca_eval",
            "unused_task_name",
            44,
            {"version": 2.0, "eval_param": "eval_param_value"},
            EvaluationBackend.ALPACA_EVAL,
            AlpacaEvalTaskParams,
        ),
        # LM Harness run with no arguments.
        (
            "lm_harness",
            "abstract_algebra",
            None,
            {},
            EvaluationBackend.LM_HARNESS,
            LMHarnessTaskParams,
        ),
        # LM Harness run with arguments.
        (
            "lm_harness",
            "abstract_algebra",
            55,
            {"num_fewshot": 44, "eval_param": "eval_param_value"},
            EvaluationBackend.LM_HARNESS,
            LMHarnessTaskParams,
        ),
        # Custom run with arguments.
        (
            "custom",
            "my_evaluation_fn",
            66,
            {"eval_param": "eval_param_value"},
            EvaluationBackend.CUSTOM,
            EvaluationTaskParams,
        ),
    ],
    ids=[
        "test_valid_initialization_alpaca_eval_no_args",
        "test_valid_initialization_alpaca_eval_with_args",
        "test_valid_initialization_lm_harness_no_args",
        "test_valid_initialization_lm_harness_with_args",
        "test_valid_initialization_custom_with_args",
    ],
)
def test_valid_initialization(
    evaluation_backend,
    task_name,
    num_samples,
    eval_kwargs,
    expected_backend,
    backend_class,
):
    # Instantiate an `EvaluationTaskParams` object.
    task_params = EvaluationTaskParams(
        evaluation_backend=evaluation_backend,
        task_name=task_name,
        num_samples=num_samples,
        eval_kwargs=eval_kwargs,
    )

    # Ensure the `EvaluationTaskParams` class members are correct.
    assert task_params.evaluation_backend == evaluation_backend
    assert task_params.get_evaluation_backend() == expected_backend
    assert task_params.task_name == task_name
    assert task_params.num_samples == num_samples
    assert task_params.eval_kwargs == eval_kwargs

    # Instantiate an `EvaluationTaskParams` subclass.
    backend_task_params = backend_class(
        evaluation_backend=evaluation_backend,
        task_name=task_name,
        num_samples=num_samples,
        eval_kwargs=eval_kwargs,
    )

    # Ensure the subclass task params are also valid.
    assert backend_task_params.evaluation_backend == evaluation_backend
    assert backend_task_params.get_evaluation_backend() == expected_backend
    assert backend_task_params.task_name == task_name
    assert backend_task_params.num_samples == num_samples
    assert backend_task_params.eval_kwargs == eval_kwargs


def test_invalid_initialization_unknown_backend():
    with pytest.raises(ValueError, match="^Unknown evaluation backend"):
        task_params = EvaluationTaskParams(
            evaluation_backend="non_existing_backend",
            task_name="some_task",
            num_samples=None,
        )
        task_params.get_evaluation_backend()


def test_invalid_initialization_no_backend():
    with pytest.raises(ValueError):
        task_params = EvaluationTaskParams(
            task_name="some_task",
            num_samples=None,
        )
        task_params.get_evaluation_backend()


@pytest.mark.parametrize(
    ("num_samples"),
    [-1, 0],
    ids=[
        "test_invalid_initialization_num_samples_negative",
        "test_invalid_initialization_num_samples_zero",
    ],
)
def test_invalid_initialization_num_samples(num_samples):
    with pytest.raises(ValueError):
        EvaluationTaskParams(
            evaluation_backend="lm_harness",
            task_name="some_task",
            num_samples=num_samples,
        )


def test_lm_harness_invalid_initialization_missing_task():
    with pytest.raises(
        ValueError, match="`task_name` must be a valid LM Harness task."
    ):
        _ = LMHarnessTaskParams(
            evaluation_backend="lm_harness",
            num_samples=10,
        )


def test_lm_harness_invalid_initialization_num_fewshot_negative():
    with pytest.raises(ValueError, match="`num_fewshot` must be non-negative."):
        _ = LMHarnessTaskParams(
            evaluation_backend="lm_harness",
            task_name="some_task",
            num_samples=10,
            num_fewshot=-1,
        )


def test_alpaca_eval_invalid_initialization_version():
    with pytest.raises(ValueError, match="AlpacaEval `version` must be 1.0 or 2.0."):
        _ = AlpacaEvalTaskParams(
            evaluation_backend="alpaca_eval",
            task_name="",
            num_samples=10,
            version=3.0,
        )



import pytest
import torch

from oumi.builders import build_data_collator, build_tokenizer
from oumi.core.configs import ModelParams
from oumi.core.constants import LABEL_IGNORE_INDEX
from oumi.core.types import ContentItem, Conversation, Message, Role
from oumi.core.types.conversation import Type


@pytest.fixture
def phi3_tokenizer():
    """Create Phi-3 tokenizer for testing."""
    model_params = ModelParams(
        model_name="microsoft/Phi-3-vision-128k-instruct",
        device_map="cpu",
        trust_remote_code=True,
        chat_template="phi3-instruct",
    )
    return build_tokenizer(model_params)


@pytest.fixture
def sample_conversation(root_testdata_dir):
    """Create a sample vision-language conversation."""
    return Conversation(
        messages=[
            Message(
                role=Role.USER,
                content=[
                    ContentItem(
                        type=Type.TEXT, content="What objects are in this image?"
                    ),
                    ContentItem(
                        type=Type.IMAGE_PATH,
                        content=str(root_testdata_dir / "images/oumi_logo_dark.png"),
                    ),
                ],
            ),
            Message(
                role=Role.ASSISTANT,
                content=[
                    ContentItem(
                        type=Type.TEXT,
                        content="This image shows the Oumi logo with stylized text.",
                    )
                ],
            ),
        ]
    )


def test_phi3_tokenization(phi3_tokenizer):
    """Test that we understand Phi-3's tokenization behavior correctly."""
    # Known tokenization from our analysis
    response_template = "<|assistant|>"
    instruction_template = "<|user|>"

    response_tokens = phi3_tokenizer.encode(response_template, add_special_tokens=False)
    instruction_tokens = phi3_tokenizer.encode(
        instruction_template, add_special_tokens=False
    )

    # Verify expected token values
    assert response_tokens == [32001], f"Expected [32001], got {response_tokens}"
    assert instruction_tokens == [32010], f"Expected [32010], got {instruction_tokens}"

    # Test full conversation tokenization
    messages = [
        {"role": "user", "content": "What is this?"},
        {"role": "assistant", "content": "This is a test."},
    ]

    prompt = phi3_tokenizer.apply_chat_template(messages, tokenize=False)
    expected_prompt = (
        "<|user|>\nWhat is this?<|end|>\n<|assistant|>\nThis is a test.<|end|>\n"
    )
    assert prompt == expected_prompt, (
        f"Expected {repr(expected_prompt)}, got {repr(prompt)}"
    )

    tokens = phi3_tokenizer.encode(prompt, add_special_tokens=False)
    expected_tokens = [
        32010,
        29871,
        13,
        5618,
        338,
        445,
        29973,
        32007,
        29871,
        13,
        32001,
        910,
        338,
        263,
        1243,
        29889,
        32007,
        29871,
        13,
    ]
    assert tokens == expected_tokens, f"Expected {expected_tokens}, got {tokens}"


def test_vision_language_completions_only(phi3_tokenizer, sample_conversation):
    """Test vision language collator with exact token-level validation."""
    # Create collator with completions-only training
    collator = build_data_collator(
        collator_name="vision_language_sft",
        tokenizer=phi3_tokenizer,
        processor_name="microsoft/Phi-3-vision-128k-instruct",
        train_on_completions_only=True,
        response_template="<|assistant|>",
        instruction_template="<|user|>",
        trust_remote_code=True,
        max_length=512,
    )

    # Process batch
    batch = [{"conversation_json": sample_conversation.to_json()}]
    result = collator(batch)

    # Verify result structure
    required_keys = ["input_ids", "labels", "attention_mask", "pixel_values"]
    for key in required_keys:
        assert key in result, f"Missing key: {key}"
        assert isinstance(result[key], torch.Tensor), f"{key} is not a tensor"

    # Extract tensors
    input_ids = result["input_ids"][0]
    labels = result["labels"][0]
    attention_mask = result["attention_mask"][0]

    # Verify shapes match
    assert input_ids.shape == labels.shape == attention_mask.shape

    # Find template positions in the tokenized sequence
    input_ids_list = input_ids.tolist()
    labels_list = labels.tolist()

    # Find assistant template position (token 32001)
    assistant_positions = [
        i for i, token in enumerate(input_ids_list) if token == 32001
    ]
    assert len(assistant_positions) == 1, (
        f"Expected 1 assistant template, found {len(assistant_positions)}"
    )

    assistant_pos = assistant_positions[0]

    # Validate masking pattern
    # Everything up to and including the assistant template should be masked
    for i in range(assistant_pos + 1):
        assert labels_list[i] == LABEL_IGNORE_INDEX, (
            f"Token at position {i} (value: {input_ids_list[i]}) should be masked, "
            f"but label is {labels_list[i]}"
        )

    # At least some tokens after the assistant template should be unmasked
    unmasked_count = sum(
        1
        for i in range(assistant_pos + 1, len(labels_list))
        if labels_list[i] != LABEL_IGNORE_INDEX
    )
    assert unmasked_count > 0, "No assistant response tokens are unmasked"

    # Verify that unmasked labels match input_ids
    for i in range(len(labels_list)):
        if labels_list[i] != LABEL_IGNORE_INDEX:
            assert labels_list[i] == input_ids_list[i], (
                f"Unmasked label at position {i} ({labels_list[i]}) "
                f"doesn't match input_id ({input_ids_list[i]})"
            )


def test_vision_language_without_completions_only(phi3_tokenizer, sample_conversation):
    # Create collator without completions-only training
    collator = build_data_collator(
        collator_name="vision_language_sft",
        tokenizer=phi3_tokenizer,
        processor_name="microsoft/Phi-3-vision-128k-instruct",
        train_on_completions_only=False,
        trust_remote_code=True,
        max_length=512,
    )

    # Process batch
    batch = [{"conversation_json": sample_conversation.to_json()}]
    result = collator(batch)

    # Verify result structure
    assert "input_ids" in result
    assert "labels" in result

    input_ids = result["input_ids"][0].tolist()
    labels = result["labels"][0].tolist()

    # Verify that input_ids and labels have the same shape
    assert len(input_ids) == len(labels)

    # Verify that unmasked labels match input_ids
    # OR image tokens are masked
    for i, (input_id, label) in enumerate(zip(input_ids, labels)):
        if label != LABEL_IGNORE_INDEX:
            assert label == input_id or label == 32000 and input_id == -1, (
                f"Position {i}: unmasked label {label} doesn't match "
                f"input_id {input_id}"
            )


def test_vision_language_completions_only_wrong_template(
    phi3_tokenizer, sample_conversation
):
    """Test exact behavior when response template is not found."""
    # Create collator with a non-existent response template
    collator = build_data_collator(
        collator_name="vision_language_sft",
        tokenizer=phi3_tokenizer,
        processor_name="microsoft/Phi-3-vision-128k-instruct",
        train_on_completions_only=True,
        response_template="<|nonexistent|>",  # This template won't be found
        trust_remote_code=True,
        max_length=512,
    )

    # Process batch
    batch = [{"conversation_json": sample_conversation.to_json()}]
    result = collator(batch)

    # When template is not found, ALL tokens should be masked
    labels = result["labels"][0].tolist()
    input_ids = result["input_ids"][0].tolist()

    # Verify every single token is masked
    for i, label in enumerate(labels):
        assert label == LABEL_IGNORE_INDEX, (
            f"Position {i} (token {input_ids[i]}) should be masked when "
            f"template not found, but label is {label}"
        )

    # Verify the count
    masked_count = sum(1 for label in labels if label == LABEL_IGNORE_INDEX)
    assert masked_count == len(labels), (
        f"Expected all {len(labels)} tokens to be masked, but only "
        f"{masked_count} are masked"
    )



# Copyright 2024 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Derived from https://github.com/volcengine/verl/blob/main/examples/data_preprocess/gsm8k.py.

This file was slightly modified to inherit from the `BaseExperimentalGrpoDataset` class.
"""

import re

import pandas as pd
from typing_extensions import override

from oumi.core.datasets.base_grpo_dataset import BaseExperimentalGrpoDataset
from oumi.core.registry import register_dataset
from oumi.core.types.conversation import Conversation


def _extract_solution(solution_str):
    """Extract the solution from the response."""
    solution = re.search("#### (\\-?[0-9\\.\\,]+)", solution_str)
    assert solution is not None
    final_solution = solution.group(0)
    final_solution = final_solution.split("#### ")[1].replace(",", "")
    return final_solution


@register_dataset("openai/gsm8k")
class Gsm8kGrpoDataset(BaseExperimentalGrpoDataset):
    """Dataset class for the `openai/gsm8k` dataset.

    A sample from the dataset::

        {
            "question": "Natalia sold clips to 48 of her friends in April, and then she sold half as many clips in May. How many clips did Natalia sell altogether in April and May?",
            "answer": "Natalia sold 48/2 = <<48/2=24>>24 clips in May.
                       Natalia sold 48+24 = <<48+24=72>>72 clips altogether in April and May.
                       #### 72"
        }
    """  # noqa: E501

    default_dataset = "openai/gsm8k"

    @override
    def transform(self, sample: pd.Series) -> dict:
        """Validate and transform the sample into Python `dict`."""
        instruction = (
            'Let\'s think step by step and output the final answer after "####".'
        )
        prompt = f"{sample['question']} {instruction}"
        solution = _extract_solution(sample["answer"])
        return {
            "data_source": "gsm8k",
            "prompt": [
                {
                    "role": "user",
                    "content": prompt,
                }
            ],
            "ability": "math",
            "reward_model": {"style": "rule", "ground_truth": solution},
            "extra_info": {
                "split": self.split if self.split else "",
                "answer": sample["answer"],
                "question": sample["question"],
            },
        }

    @override
    def transform_conversation(self, sample: pd.Series) -> Conversation:
        """Validate and transform the sample into Python `dict`."""
        sample_dict = self.transform(sample)
        prompt_message = sample_dict["prompt"][0]
        del sample_dict["prompt"]
        conversation_dict = {"messages": [prompt_message], "metadata": sample_dict}
        return Conversation.from_dict(conversation_dict)



from unittest.mock import Mock, call, patch

import pytest

from oumi.core.configs import JobConfig, JobResources, StorageMount
from oumi.core.launcher import JobState, JobStatus
from oumi.core.registry import REGISTRY, RegistryType
from oumi.launcher.clients.slurm_client import SlurmClient
from oumi.launcher.clouds.slurm_cloud import SlurmCloud
from oumi.launcher.clusters.slurm_cluster import SlurmCluster


#
# Fixtures
#
@pytest.fixture
def mock_slurm_client():
    with patch("oumi.launcher.clouds.slurm_cloud.SlurmClient") as client:
        yield client


@pytest.fixture
def mock_slurm_cluster():
    with patch("oumi.launcher.clouds.slurm_cloud.SlurmCluster") as cluster:
        cluster.get_slurm_connections.return_value = []
        cluster.parse_cluster_name = SlurmCluster.parse_cluster_name
        yield cluster


@pytest.fixture
def mock_get_slurm_connections():
    with patch(
        "oumi.launcher.clouds.slurm_cloud.SlurmCluster.get_slurm_connections"
    ) as get_conns:
        get_conns.return_value = []
        yield get_conns


@pytest.fixture
def mock_parse_cluster_name():
    with patch(
        "oumi.launcher.clouds.slurm_cloud.SlurmCluster.parse_cluster_name"
    ) as parse_name:
        yield parse_name


def _get_default_job(cloud: str) -> JobConfig:
    resources = JobResources(
        cloud=cloud,
        region="us-central1",
        zone=None,
        accelerators="A100-80GB",
        cpus="4",
        memory="64",
        instance_type=None,
        use_spot=True,
        disk_size=512,
        disk_tier="low",
    )
    return JobConfig(
        name="myjob",
        user="user",
        working_dir="./",
        num_nodes=2,
        resources=resources,
        envs={"var1": "val1"},
        file_mounts={},
        storage_mounts={
            "~/home/remote/path/gcs/": StorageMount(
                source="gs://mybucket/", store="gcs"
            )
        },
        setup="pip install -r requirements.txt",
        run="./hello_world.sh",
    )


#
# Tests
#
def test_slurm_cloud_up_cluster(mock_slurm_client, mock_slurm_cluster):
    cloud = SlurmCloud()
    mock_client = Mock(spec=SlurmClient)
    mock_slurm_client.side_effect = [mock_client]
    mock_cluster = Mock(spec=SlurmCluster)
    mock_slurm_cluster.side_effect = [mock_cluster]
    expected_job_status = JobStatus(
        id="job_id",
        cluster="user@somehost",
        name="foo",
        status="running",
        metadata="bar",
        done=False,
        state=JobState.PENDING,
    )
    mock_cluster.run_job.return_value = expected_job_status
    job = _get_default_job("slurm")
    job_status = cloud.up_cluster(job, "user@somehost")
    mock_slurm_client.assert_called_once_with(
        user="user", slurm_host="somehost", cluster_name="user@somehost"
    )
    mock_cluster.run_job.assert_called_once_with(job)
    assert job_status == expected_job_status


def test_slurm_cloud_up_cluster_fails_mismatched_user(
    mock_slurm_client, mock_slurm_cluster
):
    cloud = SlurmCloud()
    with pytest.raises(
        ValueError,
        match=(
            "Invalid cluster name: `user1@somehost`. "
            "User must match the provided job user: `user`."
        ),
    ):
        _ = cloud.up_cluster(_get_default_job("slurm"), "user1@somehost")


def test_slurm_cloud_init_with_connections(
    mock_slurm_client, mock_get_slurm_connections
):
    mock_client = Mock(spec=SlurmClient)
    mock_get_slurm_connections.return_value = [
        SlurmCluster.ConnectionInfo(user="user1", hostname="host1"),
        SlurmCluster.ConnectionInfo(user="user2", hostname="host2"),
    ]
    mock_slurm_client.side_effect = [mock_client, mock_client]
    cloud = SlurmCloud()
    cluster_names = [cluster.name() for cluster in cloud.list_clusters()]
    cluster_names.sort()
    assert cluster_names == [
        "user1@host1",
        "user2@host2",
    ]
    mock_slurm_client.assert_has_calls(
        [
            call(user="user1", slurm_host="host1", cluster_name="user1@host1"),
            call(user="user2", slurm_host="host2", cluster_name="user2@host2"),
        ]
    )


def test_slurm_cloud_init_skips_malformed_connections(
    mock_slurm_client, mock_get_slurm_connections
):
    mock_client = Mock(spec=SlurmClient)
    mock_get_slurm_connections.return_value = [
        SlurmCluster.ConnectionInfo(user="user1", hostname="host1"),
        SlurmCluster.ConnectionInfo(user="user2", hostname="host2"),
        SlurmCluster.ConnectionInfo(user="user3", hostname="host3"),
    ]
    mock_slurm_client.side_effect = [mock_client, mock_client, mock_client]
    cloud = SlurmCloud()
    cluster_names = [cluster.name() for cluster in cloud.list_clusters()]
    cluster_names.sort()
    assert cluster_names == [
        "user1@host1",
        "user2@host2",
        "user3@host3",
    ]
    mock_slurm_client.assert_has_calls(
        [
            call(user="user1", slurm_host="host1", cluster_name="user1@host1"),
            call(user="user2", slurm_host="host2", cluster_name="user2@host2"),
            call(user="user3", slurm_host="host3", cluster_name="user3@host3"),
        ]
    )


def test_slurm_cloud_initialize_cluster(mock_slurm_client, mock_get_slurm_connections):
    cloud = SlurmCloud()
    mock_get_slurm_connections.return_value = [
        SlurmCluster.ConnectionInfo(user="user1", hostname="host1"),
        SlurmCluster.ConnectionInfo(user="user2", hostname="host2"),
        SlurmCluster.ConnectionInfo(user="user3", hostname="host3"),
    ]
    mock_client = Mock(spec=SlurmClient)
    mock_slurm_client.side_effect = [mock_client, mock_client, mock_client]
    clusters = cloud.initialize_clusters()
    clusters2 = cloud.initialize_clusters()
    mock_slurm_client.assert_has_calls(
        [
            call(user="user1", slurm_host="host1", cluster_name="user1@host1"),
            call(user="user2", slurm_host="host2", cluster_name="user2@host2"),
            call(user="user3", slurm_host="host3", cluster_name="user3@host3"),
        ]
    )
    cluster_names = [cluster.name() for cluster in clusters]
    cluster_names.sort()
    assert cluster_names == [
        "user1@host1",
        "user2@host2",
        "user3@host3",
    ]
    # Verify that the second initialization returns the same clusters.
    assert clusters == clusters2


def test_slurm_cloud_list_clusters(mock_slurm_client, mock_get_slurm_connections):
    cloud = SlurmCloud()
    mock_get_slurm_connections.return_value = [
        SlurmCluster.ConnectionInfo(user="user1", hostname="host1"),
        SlurmCluster.ConnectionInfo(user="user2", hostname="host2"),
        SlurmCluster.ConnectionInfo(user="user3", hostname="host3"),
    ]
    mock_client = Mock(spec=SlurmClient)
    mock_slurm_client.side_effect = [mock_client, mock_client, mock_client]
    assert [] == cloud.list_clusters()
    clusters = cloud.initialize_clusters()
    mock_slurm_client.assert_has_calls(
        [
            call(user="user1", slurm_host="host1", cluster_name="user1@host1"),
            call(user="user2", slurm_host="host2", cluster_name="user2@host2"),
            call(user="user3", slurm_host="host3", cluster_name="user3@host3"),
        ]
    )
    clusters = cloud.list_clusters()
    expected_clusters = [
        "user1@host1",
        "user2@host2",
        "user3@host3",
    ]
    cluster_names = [cluster.name() for cluster in clusters]
    cluster_names.sort()
    assert cluster_names == expected_clusters


def test_slurm_cloud_get_cluster_empty(mock_slurm_client):
    cloud = SlurmCloud()
    # Check that there are no initial clusters.
    assert cloud.get_cluster("debug.user") is None


def test_slurm_cloud_get_cluster_success(mock_slurm_client, mock_get_slurm_connections):
    mock_client = Mock(spec=SlurmClient)
    mock_get_slurm_connections.side_effect = [
        [],
        [
            SlurmCluster.ConnectionInfo(user="user1", hostname="host1"),
            SlurmCluster.ConnectionInfo(user="user2", hostname="host2"),
            SlurmCluster.ConnectionInfo(user="user3", hostname="host3"),
        ],
    ]
    mock_slurm_client.side_effect = [mock_client, mock_client, mock_client]
    cloud = SlurmCloud()
    assert [] == cloud.list_clusters()
    _ = cloud.initialize_clusters()
    mock_slurm_client.assert_has_calls(
        [
            call(user="user1", slurm_host="host1", cluster_name="user1@host1"),
            call(user="user2", slurm_host="host2", cluster_name="user2@host2"),
            call(user="user3", slurm_host="host3", cluster_name="user3@host3"),
        ]
    )
    expected_clusters = [
        "user1@host1",
        "user2@host2",
        "user3@host3",
    ]
    for name in expected_clusters:
        cluster = cloud.get_cluster(name)
        assert cluster is not None
        assert cluster.name() == name


def test_slurm_cloud_get_cluster_fails(mock_slurm_client):
    cloud = SlurmCloud()
    mock_client = Mock(spec=SlurmClient)
    mock_slurm_client.side_effect = [mock_client, mock_client]
    cloud.initialize_clusters()
    assert cloud.get_cluster("nonexistent") is None


def test_slurm_cloud_builder_registered():
    assert REGISTRY.contains("slurm", RegistryType.CLOUD)



import pytest

from oumi.datasets.grpo.rewards import compute_letter_count_reward


@pytest.mark.parametrize(
    "s,target_count,reward",
    [
        # No valid answer
        ("foo bar 1", 1, -3),
        # Valid correct answer
        (r"\boxed{1}", 1, 0.0),
        # Valid correct answer
        (r"\boxed{+1}", 1, 0.0),
        # Valid incorrect answer
        (r"\boxed{4}", 1, -1.71429),
        # Valid incorrect answer
        (r"\boxed{-1}", 1, -1.6),
        # Invalid answer
        (r"The answer is \boxed{one}", 0, -3.0),
        # Conflicting answers
        (r"\boxed{1} \boxed{2}", 1, -3.0),
        # Valid incorrect answer
        (r"The number of 'r's in strawberry is \boxed{10}.", 3, -1.86667),
    ],
)
def test_compute_soft_target_token_length_reward(s, target_count, reward):
    calculated_reward = compute_letter_count_reward(s, target_count=target_count)
    assert calculated_reward == pytest.approx(reward, rel=1e-3)



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

from oumi.core.configs.params.evaluation_params import EvaluationBackend
from oumi.utils.packaging import PackagePrerequisites, check_package_prerequisites

# The `BACKEND_PREREQUISITES` dictionary is 2-levels deep (`dict` of nested `dict`s)
# and contains the list of prerequisites (`PackagePrerequisites`) for each evaluation
# backend. Specifically:
# - The 1st-level key determines the evaluation backend (`EvaluationBackend` Enum).
# - The 2nd-level key is an `str` that signifies either:
#   - The task name of the task to be executed in the backend.
#   - The key `ALL_TASK_PREREQUISITES_KEY`, which returns the aggregate backend
#     package prerequisites, applicable to every task that is executed.
ALL_TASK_PREREQUISITES_KEY = "all_task_prerequisites"
BACKEND_PREREQUISITES: dict[
    EvaluationBackend, dict[str, list[PackagePrerequisites]]
] = {
    EvaluationBackend.LM_HARNESS: {
        ALL_TASK_PREREQUISITES_KEY: [],
        "leaderboard_ifeval": [
            PackagePrerequisites("langdetect"),
            PackagePrerequisites("immutabledict"),
            PackagePrerequisites("nltk", "3.9.1"),
        ],
        "leaderboard_math_hard": [
            PackagePrerequisites("antlr4-python3-runtime", "4.11", "4.11"),
            PackagePrerequisites("sympy", "1.12"),
            PackagePrerequisites("sentencepiece", "0.1.98"),
        ],
    },
    EvaluationBackend.ALPACA_EVAL: {
        ALL_TASK_PREREQUISITES_KEY: [PackagePrerequisites("alpaca_eval")]
    },
    EvaluationBackend.CUSTOM: {ALL_TASK_PREREQUISITES_KEY: []},
}


def check_prerequisites(
    evaluation_backend: EvaluationBackend,
    task_name: Optional[str] = None,
) -> None:
    """Check whether the evaluation backend prerequisites are satisfied.

    Args:
        evaluation_backend: The evaluation backend that the task will run.
        task_name (for LM Harness backend only): The name of the task to run.

    Raises:
        RuntimeError: If the evaluation backend prerequisites are not satisfied.
    """
    # Error message prefixes and suffixes.
    task_reference = f"({task_name}) " if task_name else ""
    runtime_error_prefix = (
        "The current evaluation cannot be launched because the "
        f"{evaluation_backend.value} backend prerequisites for the specific task "
        f"{task_reference}are not satisfied. In order to proceed, the following "
        "package(s) must be installed and have the correct version:\n"
    )
    runtime_error_suffix = (
        "\nNote that you can install all evaluation-related packages with the "
        "following command:\n`pip install oumi[evaluation]`"
    )

    # Per backend prerequisite checks.
    backend_prerequisites_dict = BACKEND_PREREQUISITES[evaluation_backend]
    package_prerequisites_list = backend_prerequisites_dict[ALL_TASK_PREREQUISITES_KEY]
    if task_name and task_name in backend_prerequisites_dict:
        package_prerequisites_list.extend(backend_prerequisites_dict[task_name])
    check_package_prerequisites(
        package_prerequisites=package_prerequisites_list,
        runtime_error_prefix=runtime_error_prefix,
        runtime_error_suffix=runtime_error_suffix,
    )



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Calls `profiler.step()`  at the end of each training step."""

import sys
from typing import Optional, Union

import torch
import transformers

from oumi.core.callbacks.base_trainer_callback import BaseTrainerCallback
from oumi.core.configs import TrainingParams


class ProfilerStepCallback(BaseTrainerCallback):
    """Trainer callback to notify PyTorch profiler about training steps completion.

    Also, adds microstep function labels using `torch.profiler.record_function()`.
    """

    def __init__(self, profiler):
        """Initialize the ProfilerStepCallback.

        Args:
            profiler: PyTorch profiler object.
        """
        self._profiler = profiler
        self._microstep_function = None

    def on_step_begin(
        self,
        args: Union[transformers.TrainingArguments, TrainingParams],
        state: Optional[transformers.TrainerState] = None,
        control: Optional[transformers.TrainerControl] = None,
        **kwargs,
    ):
        """Event called at the beginning of a training step.

        If using gradient accumulation, one training step might take several inputs.
        """
        self._complete_previous_microstep_if_needed()
        self._start_microstep()

    def on_substep_end(
        self,
        args: Union[transformers.TrainingArguments, TrainingParams],
        state: Optional[transformers.TrainerState] = None,
        control: Optional[transformers.TrainerControl] = None,
        **kwargs,
    ):
        """Event called at the end of an substep during gradient accumulation."""
        self._complete_previous_microstep_if_needed()
        self._start_microstep()

    def on_step_end(
        self,
        args: Union[transformers.TrainingArguments, TrainingParams],
        state: Optional[transformers.TrainerState] = None,
        control: Optional[transformers.TrainerControl] = None,
        **kwargs,
    ):
        """Event called at the end of each train step.

        Note that this will be called after all gradient accumulation substeps.
        """
        self._complete_previous_microstep_if_needed()
        self._profiler.step()

    def _complete_previous_microstep_if_needed(self):
        if self._microstep_function is None:
            return

        self._microstep_function.__exit__(*sys.exc_info())
        self._microstep_function = None

    def _start_microstep(self):
        self._microstep_function = torch.profiler.record_function("microstep")
        self._microstep_function.__enter__()



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass, field
from typing import Optional

from oumi.core.configs.base_config import BaseConfig
from oumi.core.configs.params.model_params import ModelParams


@dataclass
class QuantizationConfig(BaseConfig):
    """Configuration for model quantization.

    Reduces model size by converting weights from higher precision (float32) to
    lower precision (int4, int8) formats while maintaining performance.

    Tested on NVIDIA H100 GPU with models up to 14B parameters.

    Example:
        >>> config = QuantizationConfig(
        ...     model=ModelParams(model_name="meta-llama/Llama-2-7b-hf"),
        ...     method="awq_q4_0",
        ...     output_path="llama2-7b-q4.gguf"
        ... )
    """

    model: ModelParams = field(default_factory=ModelParams)
    """Model to quantize. Supports HuggingFace IDs, local paths, or Oumi models."""

    method: str = "awq_q4_0"
    """Quantization method. AWQ methods (awq_q4_0, awq_q8_0) provide best quality.
    Direct GGUF methods (q4_0, q8_0) for llama.cpp. Precision methods (f16, f32)."""

    output_path: str = "quantized_model"
    """Output file path for the quantized model."""

    output_format: str = "safetensors"
    """Output format: 'safetensors'."""

    batch_size: Optional[int] = None
    """Batch size for calibration. Auto-sized if None. Typical: 32, 8-32, 1-8."""

    verbose: bool = False
    """Enable detailed progress logging."""

    # AWQ-specific configuration
    awq_group_size: int = 128
    """AWQ weight grouping size. 128 (balanced), 64 (higher accuracy), 256 (faster)."""

    awq_zero_point: bool = True
    """Enable zero-point quantization for AWQ. Generally recommended."""

    awq_version: str = "GEMM"
    """AWQ kernel version. 'GEMM' (faster, default) or 'GEMV'."""

    cleanup_temp: bool = True
    """Remove temporary AWQ files after conversion."""

    calibration_samples: int = 512
    """AWQ calibration samples. 512 (balanced), 128 (faster), 1024 (more accurate)."""

    def __post_init__(self):
        """Post-initialization validation."""
        from oumi.quantize.constants import SUPPORTED_METHODS, SUPPORTED_OUTPUT_FORMATS

        # Validate output format
        if self.output_format not in SUPPORTED_OUTPUT_FORMATS:
            raise ValueError(
                f"Unsupported output format: {self.output_format}. "
                f"Must be one of: {SUPPORTED_OUTPUT_FORMATS}."
            )

        # Validate quantization method
        if self.method not in SUPPORTED_METHODS:
            raise ValueError(
                f"Unsupported quantization method: {self.method}. "
                f"Must be one of: {SUPPORTED_METHODS}."
            )



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Vision-Language datasets module."""

from oumi.datasets.vision_language.coco_captions import COCOCaptionsDataset
from oumi.datasets.vision_language.docmatix import DocmatixDataset
from oumi.datasets.vision_language.flickr30k import Flickr30kDataset
from oumi.datasets.vision_language.geometry3k import Geometry3kDataset
from oumi.datasets.vision_language.huggingface import HuggingFaceVisionDataset
from oumi.datasets.vision_language.llava_instruct_mix_vsft import (
    LlavaInstructMixVsftDataset,
)
from oumi.datasets.vision_language.lmms_lab_multimodal_open_r1 import (
    LmmsLabMultimodalOpenR1Dataset,
)
from oumi.datasets.vision_language.mnist_sft import MnistSftDataset
from oumi.datasets.vision_language.pixmo_ask_model_anything import (
    PixmoAskModelAnythingDataset,
)
from oumi.datasets.vision_language.pixmo_cap import PixmoCapDataset
from oumi.datasets.vision_language.pixmo_cap_qa import PixmoCapQADataset
from oumi.datasets.vision_language.rlaif_v import OpenbmbRlaifVDataset
from oumi.datasets.vision_language.the_cauldron import TheCauldronDataset
from oumi.datasets.vision_language.vision_dpo_jsonlines import (
    VisionDpoJsonlinesDataset,
)
from oumi.datasets.vision_language.vision_jsonlines import VLJsonlinesDataset
from oumi.datasets.vision_language.vqav2_small import Vqav2SmallDataset

__all__ = [
    "COCOCaptionsDataset",
    "DocmatixDataset",
    "Flickr30kDataset",
    "Geometry3kDataset",
    "HuggingFaceVisionDataset",
    "LlavaInstructMixVsftDataset",
    "LmmsLabMultimodalOpenR1Dataset",
    "MnistSftDataset",
    "OpenbmbRlaifVDataset",
    "PixmoAskModelAnythingDataset",
    "PixmoCapDataset",
    "PixmoCapQADataset",
    "TheCauldronDataset",
    "VisionDpoJsonlinesDataset",
    "VLJsonlinesDataset",
    "Vqav2SmallDataset",
]



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import NamedTuple, Optional

from oumi.core.types.conversation import Conversation


class FeatureGeneratorOptions(NamedTuple):
    """Options for the feature generator."""

    allow_feature_reshape: bool = True
    """Whether to allow reshaping of the model input features.

    For example, whether the generator can drop the first dummy dimension.
    """


class BaseConversationFeatureGenerator(ABC):
    """Applies `processor` to generate model inputs from an input `Conversation`."""

    @abstractmethod
    def transform_conversation(
        self, conversation: Conversation, options: Optional[FeatureGeneratorOptions]
    ) -> dict:
        """Transforms a single Oumi conversation into a dictionary of model inputs.

        Args:
            conversation: An input conversation.
            options: Options for the feature generator.

        Returns:
            dict: A dictionary of inputs for a model.
        """
        raise NotImplementedError

    @abstractmethod
    def transform_conversations(
        self,
        conversations: list[Conversation],
        options: Optional[FeatureGeneratorOptions],
    ) -> dict:
        """Transforms a list of Oumi conversations into a dictionary of model inputs.

        Args:
            conversations: A list of input conversations.
            options: Options for the feature generator.

        Returns:
            dict: A dictionary of inputs for a model.
        """
        raise NotImplementedError



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from typing import Optional

import numpy as np

from oumi.core.configs.params.base_params import BaseParams


@dataclass
class RemoteParams(BaseParams):
    """Parameters for running inference against a remote API."""

    api_url: Optional[str] = None
    """URL of the API endpoint to use for inference."""

    api_key: Optional[str] = None
    """API key to use for authentication."""

    api_key_env_varname: Optional[str] = None
    """Name of the environment variable containing the API key for authentication."""

    max_retries: int = 3
    """Maximum number of retries to attempt when calling an API."""

    retry_backoff_base: float = 1.0
    """Base delay in seconds for exponential backoff between retries."""

    retry_backoff_max: float = 30.0
    """Maximum delay in seconds between retries."""

    connection_timeout: float = 300.0
    """Timeout in seconds for a request to an API."""

    num_workers: int = 1
    """Number of workers to use for parallel inference."""

    politeness_policy: float = 0.0
    """Politeness policy to use when calling an API.

    If greater than zero, this is the amount of time in seconds a worker will sleep
    before making a subsequent request.
    """

    batch_completion_window: Optional[str] = "24h"
    """Time window for batch completion. Currently only '24h' is supported.

    Only used for batch inference.
    """

    use_adaptive_concurrency: bool = True
    """Whether to use adaptive concurrency control.

    If True, the number of concurrent requests will be adjusted based on the error rate
    of the requests. As error rate increases above a threshold, the number of concurrent
    requests will decrease, and as error rate decreases below a threshold, the number of
    concurrent requests will increase.

    When this is enabled, users should set `num_workers` to the requests per minute
    (RPM/QPM) of the model/API, and the `politeness_policy` to 60s (as most APIs query
    limits are dictated by the number of requests per minute).

    The lowest concurrency can be is 1, and the highest concurrency is `num_workers`.
    Updates to concurrency will happen no sooner than `politeness_policy` seconds after
    the last update, and at least 10 requests must have been made since the last update.

    In the event that even 1 concurrency causes the error rate to exceed the threshold,
    it is recommended to increase the `politeness_policy` to allow more time between
    requests.
    """

    def __post_init__(self):
        """Validate the remote parameters."""
        if self.num_workers < 1:
            raise ValueError(
                "Number of num_workers must be greater than or equal to 1."
            )
        if self.politeness_policy < 0:
            raise ValueError("Politeness policy must be greater than or equal to 0.")
        if self.connection_timeout < 0:
            raise ValueError("Connection timeout must be greater than or equal to 0.")
        if not np.isfinite(self.politeness_policy):
            raise ValueError("Politeness policy must be finite.")
        if self.max_retries < 0:
            raise ValueError("Max retries must be greater than or equal to 0.")
        if self.retry_backoff_base <= 0:
            raise ValueError("Retry backoff base must be greater than 0.")
        if self.retry_backoff_max < self.retry_backoff_base:
            raise ValueError(
                "Retry backoff max must be greater than or equal to retry backoff base."
            )


@dataclass
class AdaptiveConcurrencyParams(BaseParams):
    """Configuration for adaptive concurrency control."""

    min_concurrency: int = 5
    """Minimum number of concurrent requests to allow.

    Backoff throttling will never reduce concurrency below this value.
    """

    max_concurrency: int = 100
    """Maximum number of concurrent requests allowed.

    The concurrency will never be allowed to go above this value.
    """

    initial_concurrency_factor: float = 0.5
    """Initial concurrency factor.

    The initial concurrency will be set to (max_concurrency - min_concurrency) *
    initial_concurrency_factor + min_concurrency.

    Example:
    - min_concurrency = 5
    - max_concurrency = 100
    - initial_concurrency_factor = 0.5
    - initial_concurrency = (100 - 5) * 0.5 + 5 = 52.5 = 52
    """

    concurrency_step: int = 5
    """How much to increase concurrency during warmup.

    During warmup, concurrency will be increased by this amount. (i.e. if concurrency is
    50, and the concurrency step is 5, the concurrency will be increased to 55). This
    change will happen no sooner than min_update_time seconds after the last update.
    """

    min_update_time: float = 60.0
    """Minimum seconds between attempted updates.

    The concurrency will not be adjusted sooner than this time.
    """

    error_threshold: float = 0.01
    """Error rate threshold (0.01 = 1%) to trigger backoff.

    If the error rate is greater than this threshold, backoff will be triggered.

    Consider keeping this value low. Once a particular error rate is reached, there are
    already other requests in-flight which will likely fail, so the sooner concurrency
    is reduced, the better chance the system has of recovering.
    """

    backoff_factor: float = 0.8
    """Factor to multiply concurrency by during backoff.

    During backoff, the concurrency will be reduced by this factor. (i.e. if concurrency
    is 100, and the backoff factor is 0.8, the concurrency will be reduced to 80).
    """

    recovery_threshold: float = 0.00
    """Error rate threshold (0.00 = 0%) to allow recovery.

    If the error rate is less than this threshold, recovery will be triggered.
    """

    min_window_size: int = 10
    """Minimum number of recent requests to consider for error rate calculation.

    If the number of requests since the last update is less than this threshold, the
    concurrency will not be adjusted.
    """

    def __post_init__(self):
        """Validate the adaptive concurrency parameters."""
        if self.min_concurrency < 1:
            raise ValueError("Min concurrency must be greater than or equal to 1.")
        if self.max_concurrency < self.min_concurrency:
            raise ValueError(
                "Max concurrency must be greater than or equal to min concurrency."
            )
        if self.initial_concurrency_factor < 0 or self.initial_concurrency_factor > 1:
            raise ValueError("Initial concurrency factor must be between 0 and 1.")
        if self.concurrency_step < 1:
            raise ValueError("Concurrency step must be greater than or equal to 1.")
        if self.min_update_time <= 0:
            raise ValueError("Min update time must be greater than 0.")
        if self.error_threshold < 0 or self.error_threshold > 1:
            raise ValueError("Error threshold must be between 0 and 1.")
        if self.backoff_factor <= 0:
            raise ValueError("Backoff factor must be greater than 0.")
        if self.recovery_threshold < 0 or self.recovery_threshold > 1:
            raise ValueError("Recovery threshold must be between 0 and 1.")
        if self.recovery_threshold >= self.error_threshold:
            raise ValueError("Recovery threshold must be less than error threshold.")
        if self.min_window_size < 1:
            raise ValueError("Min window size must be greater than or equal to 1.")



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from pathlib import Path
from typing import Optional

import pandas as pd
from datasets import load_dataset

from oumi.core.configs.params.synthesis_params import DatasetSource


class DatasetStorageType(Enum):
    """Storage location for a dataset (local, HuggingFace, etc.)."""

    HF = "hf"
    """HuggingFace"""

    OUMI = "oumi"
    """Oumi"""

    LOCAL = "local"
    """Local files"""


class DatasetPath:
    """Path to a dataset in some storage location."""

    def __init__(self, path: str):
        """Initialize the dataset path.

        Args:
            path: The path to the dataset.

        Possible path formats:
        HuggingFace:
        - "hf:repo_id/dataset_name"
        Oumi:
        - "oumi:dataset_id"
        Local:
        - "path/to/data/file.jsonl"
        - "path/to/data/file.csv"
        - "path/to/data/file.tsv"
        - "path/to/data/file.parquet"
        - "path/to/data/file.json"
        - "path/to/data/*.jsonl"
        - "path/to/data/*.csv"
        - "path/to/data/*.tsv"
        - "path/to/data/*.parquet"
        - "path/to/data/*.json"
        """
        self._path = path
        self._storage_type = self._get_storage_type(path)
        self._file_extension = ""
        if self._storage_type == DatasetStorageType.LOCAL:
            self._file_extension = self._get_file_extension(path)
            if self._file_extension not in ["jsonl", "csv", "tsv", "parquet", "json"]:
                raise ValueError(f"Invalid path: {path}")

    def _get_storage_type(self, path: str) -> DatasetStorageType:
        """Get the storage type from the path."""
        prefix = path.split(":")[0]
        if prefix == "hf":
            return DatasetStorageType.HF
        elif prefix == "oumi":
            return DatasetStorageType.OUMI
        else:
            return DatasetStorageType.LOCAL

    def _get_file_extension(self, path: str) -> str:
        """Get the file extension from the path."""
        if self._storage_type == DatasetStorageType.LOCAL:
            return Path(path).suffix.lower()[1:]
        else:
            raise NotImplementedError(
                f"No extension for {self._storage_type} storage type"
            )

    def get_path_str(self) -> str:
        """Get the path."""
        if (
            self._storage_type == DatasetStorageType.HF
            or self._storage_type == DatasetStorageType.OUMI
        ):
            return self._path.split(":")[1]
        else:
            return self._path

    def get_storage_type(self) -> DatasetStorageType:
        """Get the storage type."""
        return self._storage_type

    def get_file_extension(self) -> str:
        """Get the file extension."""
        if self._storage_type == DatasetStorageType.LOCAL:
            return self._file_extension
        else:
            raise NotImplementedError(
                f"No extension for {self._storage_type} storage type"
            )


class DatasetReader:
    """Reads a dataset from some storage location.

    Supports:
    - HuggingFace
    - Local files (JSONL, CSV, TSV, Parquet, JSON)
    - Glob patterns
    """

    def read(self, data_source: DatasetSource) -> list[dict]:
        """Read the data from the data path."""
        data_path = DatasetPath(data_source.path)
        storage_type = data_path.get_storage_type()
        if storage_type == DatasetStorageType.HF:
            samples = self._read_from_hf(
                data_path.get_path_str(),
                data_source.hf_split,
                data_source.hf_revision,
            )
        elif storage_type == DatasetStorageType.LOCAL:
            file_extension = data_path.get_file_extension()
            if "*" in data_path.get_path_str():
                samples = self._read_from_glob(data_path.get_path_str(), file_extension)
            else:
                samples = self._read_from_local(
                    data_path.get_path_str(),
                    file_extension,
                )
        elif storage_type == DatasetStorageType.OUMI:
            raise NotImplementedError(
                "Oumi storage type is not supported in open source."
            )
        else:
            raise ValueError(f"Unsupported storage type: {storage_type}")

        if data_source.attribute_map:
            samples = self._map_attributes(samples, data_source.attribute_map)

        return samples

    def _map_attributes(
        self,
        samples: list[dict],
        attribute_map: dict[str, str],
    ) -> list[dict]:
        """Map the attributes to the samples."""
        new_samples = []
        for sample in samples:
            new_sample = {}
            for old_key, new_key in attribute_map.items():
                new_sample[new_key] = sample[old_key]
            new_samples.append(new_sample)
        return new_samples

    def _read_from_hf(
        self,
        hf_path: str,
        split: Optional[str] = None,
        revision: Optional[str] = None,
    ) -> list[dict]:
        """Read the dataset from HuggingFace."""
        dataset = load_dataset(hf_path, split=split, revision=revision)
        samples = []
        for sample in dataset:
            if not isinstance(sample, dict):
                raise ValueError(f"Expected dict, got {type(sample)}")
            samples.append(sample)
        return samples

    def _read_from_local(self, local_path: str, file_extension: str) -> list[dict]:
        """Read the dataset from the local path."""
        if file_extension == "jsonl":
            return self._read_from_jsonl(local_path)
        elif file_extension == "csv":
            return self._read_from_csv(local_path, sep=",")
        elif file_extension == "tsv":
            return self._read_from_csv(local_path, sep="\t")
        elif file_extension == "parquet":
            return self._read_from_parquet(local_path)
        elif file_extension == "json":
            return self._read_from_json(local_path)
        else:
            raise ValueError(f"Unsupported local path suffix: {file_extension}")

    def _read_from_glob(self, glob_path: str, file_extension: str) -> list[dict]:
        """Read the data from the glob path."""
        path = Path(glob_path)
        files = Path(path.parent).glob(path.name)
        samples = []
        for file in files:
            samples.extend(self._read_from_local(file.as_posix(), file_extension))
        return samples

    def _read_from_jsonl(self, jsonl_path: str) -> list[dict]:
        """Read the dataset from a JSONL file."""
        jsonl_df = pd.read_json(jsonl_path, lines=True)
        return jsonl_df.to_dict(orient="records")

    def _read_from_csv(self, csv_path: str, sep: str = ",") -> list[dict]:
        """Read the dataset from a CSV file."""
        csv_df = pd.read_csv(csv_path, sep=sep)
        return csv_df.to_dict(orient="records")

    def _read_from_parquet(self, parquet_path: str) -> list[dict]:
        """Read the dataset from a Parquet file."""
        parquet_df = pd.read_parquet(parquet_path)
        return parquet_df.to_dict(orient="records")

    def _read_from_json(self, json_path: str) -> list[dict]:
        """Read the dataset from a JSON file."""
        json_df = pd.read_json(json_path)
        return json_df.to_dict(orient="records")
